from __future__ import with_statement

import sys
import unittest

from nose import SkipTest
from nose.tools import assert_raises, assert_equal
from pycassa import ConnectionPool, ColumnFamily, NotFoundException
import pycassa.batch as batch_mod
from pycassa.system_manager import SystemManager

ROWS = {'1': {'a': '123', 'b': '123'},
        '2': {'a': '234', 'b': '234'},
        '3': {'a': '345', 'b': '345'}}

pool = cf = scf = counter_cf = super_counter_cf = sysman = None

def setup_module():
    global pool, cf, scf, counter_cf, super_counter_cf, sysman
    credentials = {'username': 'jsmith', 'password': 'havebadpass'}
    pool = ConnectionPool(keyspace='PycassaTestKeyspace', credentials=credentials)
    cf = ColumnFamily(pool, 'Standard1')
    scf = ColumnFamily(pool, 'Super1')
    sysman = SystemManager()
    counter_cf = ColumnFamily(pool, 'Counter1')
    super_counter_cf = ColumnFamily(pool, 'SuperCounter1')

def teardown_module():
    pool.dispose()

class TestMutator(unittest.TestCase):

    def tearDown(self):
        for key, cols in cf.get_range():
            cf.remove(key)
        for key, cols, in scf.get_range():
            scf.remove(key)

    def test_insert(self):
        batch = cf.batch()
        for key, cols in ROWS.iteritems():
            batch.insert(key, cols)
        batch.send()
        for key, cols in ROWS.items():
            assert cf.get(key) == cols

    def test_insert_supercolumns(self):
        batch = scf.batch()
        batch.insert('one', ROWS)
        batch.insert('two', ROWS)
        batch.insert('three', ROWS)
        batch.send()
        assert scf.get('one') == ROWS
        assert scf.get('two') == ROWS
        assert scf.get('three') == ROWS

    def test_insert_counters(self):
        batch = counter_cf.batch()
        batch.insert('one', {'col': 1})
        batch.insert('two', {'col': 2})
        batch.insert('three', {'col': 3})
        batch.send()
        assert_equal(counter_cf.get('one'), {'col': 1})
        assert_equal(counter_cf.get('two'), {'col': 2})
        assert_equal(counter_cf.get('three'), {'col': 3})

        batch = super_counter_cf.batch()
        batch.insert('one', {'scol': {'col1': 1, 'col2': 2}})
        batch.insert('two', {'scol': {'col1': 3, 'col2': 4}})
        batch.send()
        assert_equal(super_counter_cf.get('one'), {'scol': {'col1': 1, 'col2': 2}})
        assert_equal(super_counter_cf.get('two'), {'scol': {'col1': 3, 'col2': 4}})

    def test_queue_size(self):
        batch = cf.batch(queue_size=2)
        batch.insert('1', ROWS['1'])
        batch.insert('2', ROWS['2'])
        batch.insert('3', ROWS['3'])
        assert cf.get('1') == ROWS['1']
        assert_raises(NotFoundException, cf.get, '3')
        batch.send()
        for key, cols in ROWS.items():
            assert cf.get(key) == cols

    def test_remove_key(self):
        batch = cf.batch()
        batch.insert('1', ROWS['1'])
        batch.remove('1')
        batch.send()
        assert_raises(NotFoundException, cf.get, '1')

    def test_remove_columns(self):
        batch = cf.batch()
        batch.insert('1', {'a': '123', 'b': '123'})
        batch.remove('1', ['a'])
        batch.send()
        assert cf.get('1') == {'b': '123'}

    def test_remove_supercolumns(self):
        batch = scf.batch()
        batch.insert('one', ROWS)
        batch.insert('two', ROWS)
        batch.insert('three', ROWS)
        batch.remove('two', ['b'], '2')
        batch.send()
        assert scf.get('one') == ROWS
        assert scf.get('two')['2'] == {'a': '234'}
        assert scf.get('three') == ROWS

    def test_chained(self):
        batch = cf.batch()
        batch.insert('1', ROWS['1']).insert('2', ROWS['2']).insert('3', ROWS['3']).send()
        assert cf.get('1') == ROWS['1']
        assert cf.get('2') == ROWS['2']
        assert cf.get('3') == ROWS['3']

    def test_contextmgr(self):
        if sys.version_info < (2, 5):
            raise SkipTest("No context managers in Python < 2.5")
        exec """with cf.batch(queue_size=2) as b:
    b.insert('1', ROWS['1'])
    b.insert('2', ROWS['2'])
    b.insert('3', ROWS['3'])
assert cf.get('3') == ROWS['3']"""

    def test_multi_column_family(self):
        batch = batch_mod.Mutator(pool)
        cf2 = cf
        batch.insert(cf, '1', ROWS['1'])
        batch.insert(cf, '2', ROWS['2'])
        batch.remove(cf2, '1', ROWS['1'])
        batch.send()
        assert cf.get('2') == ROWS['2']
        assert_raises(NotFoundException, cf.get, '1')

    def test_atomic_insert_at_mutator_creation(self):
        batch = cf.batch(atomic=True)
        for key, cols in ROWS.iteritems():
            batch.insert(key, cols)
        batch.send()
        for key, cols in ROWS.items():
            assert cf.get(key) == cols

    def test_atomic_insert_at_send(self):
        batch = cf.batch(atomic=True)
        for key, cols in ROWS.iteritems():
            batch.insert(key, cols)
        batch.send(atomic=True)
        for key, cols in ROWS.items():
            assert cf.get(key) == cols



import unittest

from nose.tools import assert_raises, assert_equal, assert_true

from pycassa import index, ColumnFamily, ConnectionPool,\
                    NotFoundException, SystemManager
from pycassa.util import OrderedDict

from tests.util import requireOPP

pool = cf = scf = indexed_cf = counter_cf = counter_scf = sys_man = None

def setup_module():
    global pool, cf, scf, indexed_cf, counter_cf, counter_scf, sys_man
    credentials = {'username': 'jsmith', 'password': 'havebadpass'}
    pool = ConnectionPool(keyspace='PycassaTestKeyspace',
            credentials=credentials, timeout=1.0)
    cf = ColumnFamily(pool, 'Standard1', dict_class=TestDict)
    scf = ColumnFamily(pool, 'Super1', dict_class=dict)
    indexed_cf = ColumnFamily(pool, 'Indexed1')
    sys_man = SystemManager()
    counter_cf = ColumnFamily(pool, 'Counter1')
    counter_scf = ColumnFamily(pool, 'SuperCounter1')

def teardown_module():
    cf.truncate()
    indexed_cf.truncate()
    counter_cf.truncate()
    counter_scf.truncate()
    pool.dispose()


class TestDict(dict):
    pass


class TestColumnFamily(unittest.TestCase):

    def setUp(self):
        self.sys_man = sys_man

    def tearDown(self):
        for key, columns in cf.get_range():
            cf.remove(key)
        for key, columns in indexed_cf.get_range():
            cf.remove(key)

    def test_bad_kwarg(self):
        assert_raises(TypeError,
                ColumnFamily.__init__, pool, 'test', bar='foo')

    def test_empty(self):
        key = 'TestColumnFamily.test_empty'
        assert_raises(NotFoundException, cf.get, key)
        assert_equal(len(cf.multiget([key])), 0)
        for key, columns in cf.get_range():
            assert_equal(len(columns), 0)

    def test_insert_get(self):
        key = 'TestColumnFamily.test_insert_get'
        columns = {'1': 'val1', '2': 'val2'}
        assert_raises(NotFoundException, cf.get, key)
        ts = cf.insert(key, columns)
        assert_true(isinstance(ts, (int, long)))
        assert_equal(cf.get(key), columns)

    def test_insert_multiget(self):
        key1 = 'TestColumnFamily.test_insert_multiget1'
        columns1 = {'1': 'val1', '2': 'val2'}
        key2 = 'test_insert_multiget1'
        columns2 = {'3': 'val1', '4': 'val2'}
        missing_key = 'key3'

        cf.insert(key1, columns1)
        cf.insert(key2, columns2)
        rows = cf.multiget([key1, key2, missing_key])
        assert_equal(len(rows), 2)
        assert_equal(rows[key1], columns1)
        assert_equal(rows[key2], columns2)
        assert_true(missing_key not in rows)

    def test_multiget_multiple_bad_key(self):
        key = 'efefefef'
        cf.multiget([key, key, key])

    def test_insert_get_count(self):
        key = 'TestColumnFamily.test_insert_get_count'
        columns = {'1': 'val1', '2': 'val2'}
        cf.insert(key, columns)
        assert_equal(cf.get_count(key), 2)

        assert_equal(cf.get_count(key, column_start='1'), 2)
        assert_equal(cf.get_count(key, column_finish='2'), 2)
        assert_equal(cf.get_count(key, column_start='1', column_finish='2'), 2)
        assert_equal(cf.get_count(key, column_start='1', column_finish='1'), 1)
        assert_equal(cf.get_count(key, columns=['1', '2']), 2)
        assert_equal(cf.get_count(key, columns=['1']), 1)
        assert_equal(cf.get_count(key, max_count=1), 1)
        assert_equal(cf.get_count(key, max_count=1, column_reversed=True), 1)
        assert_equal(cf.get_count(key, column_reversed=True), 2)
        assert_equal(cf.get_count(key, column_start='1', column_reversed=True), 1)

    def test_insert_multiget_count(self):
        keys = ['TestColumnFamily.test_insert_multiget_count1',
                'TestColumnFamily.test_insert_multiget_count2',
                'TestColumnFamily.test_insert_multiget_count3']
        for key in keys:
            cf.insert(key, {'1': 'val1', '2': 'val2'})

        result = cf.multiget_count(keys)
        assert_equal([result[k] for k in keys], [2 for key in keys])

        result = cf.multiget_count(keys, column_start='1')
        assert_equal([result[k] for k in keys], [2 for key in keys])

        result = cf.multiget_count(keys, column_finish='2')
        assert_equal([result[k] for k in keys], [2 for key in keys])

        result = cf.multiget_count(keys, column_start='1', column_finish='2')
        assert_equal([result[k] for k in keys], [2 for key in keys])

        result = cf.multiget_count(keys, column_start='1', column_finish='1')
        assert_equal([result[k] for k in keys], [1 for key in keys])

        result = cf.multiget_count(keys, columns=['1', '2'])
        assert_equal([result[k] for k in keys], [2 for key in keys])

        result = cf.multiget_count(keys, columns=['1'])
        assert_equal([result[k] for k in keys], [1 for key in keys])

        result = cf.multiget_count(keys, max_count=1)
        assert_equal([result[k] for k in keys], [1 for key in keys])

        result = cf.multiget_count(keys, column_start='1', column_reversed=True)
        assert_equal([result[k] for k in keys], [1 for key in keys])

    @requireOPP
    def test_insert_get_range(self):
        keys = ['TestColumnFamily.test_insert_get_range%s' % i for i in xrange(5)]
        columns = {'1': 'val1', '2': 'val2'}
        for key in keys:
            cf.insert(key, columns)

        rows = list(cf.get_range(start=keys[0], finish=keys[-1]))
        assert_equal(len(rows), len(keys))
        for i, (k, c) in enumerate(rows):
            assert_equal(k, keys[i])
            assert_equal(c, columns)

    @requireOPP
    def test_get_range_batching(self):
        cf.truncate()

        keys = []
        columns = {'c': 'v'}
        for i in range(100, 201):
            keys.append('key%d' % i)
            cf.insert('key%d' % i, columns)

        for i in range(201, 301):
            cf.insert('key%d' % i, columns)

        count = 0
        for k, v in cf.get_range(row_count=100, buffer_size=10):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 100)

        count = 0
        for k, v in cf.get_range(row_count=100, buffer_size=1000):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 100)

        count = 0
        for k, v in cf.get_range(row_count=100, buffer_size=150):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 100)

        count = 0
        for k, v in cf.get_range(row_count=100, buffer_size=7):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 100)

        count = 0
        for k, v in cf.get_range(row_count=100, buffer_size=2):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 100)

        # Put the remaining keys in our list
        for i in range(201, 301):
            keys.append('key%d' % i)

        count = 0
        for k, v in cf.get_range(row_count=10000, buffer_size=2):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        count = 0
        for k, v in cf.get_range(row_count=10000, buffer_size=7):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        count = 0
        for k, v in cf.get_range(row_count=10000, buffer_size=200):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        count = 0
        for k, v in cf.get_range(row_count=10000, buffer_size=10000):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        # Don't give a row count
        count = 0
        for k, v in cf.get_range(buffer_size=2):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        count = 0
        for k, v in cf.get_range(buffer_size=77):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        count = 0
        for k, v in cf.get_range(buffer_size=200):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        count = 0
        for k, v in cf.get_range(buffer_size=10000):
            assert_true(k in keys, 'key "%s" should be in keys' % k)
            count += 1
        assert_equal(count, 201)

        cf.truncate()

    @requireOPP
    def test_get_range_tokens(self):
        cf.truncate()
        columns = {'c': 'v'}
        for i in range(100, 201):
            cf.insert('key%d' % i, columns)

        results = list(cf.get_range(start_token="key100".encode('hex'), finish_token="key200".encode('hex')))
        assert_equal(100, len(results))

        results = list(cf.get_range(start_token="key100".encode('hex'), finish_token="key200".encode('hex'), buffer_size=10))
        assert_equal(100, len(results))

        results = list(cf.get_range(start_token="key100".encode('hex'), buffer_size=10))
        assert_equal(100, len(results))

        results = list(cf.get_range(finish_token="key201".encode('hex'), buffer_size=10))
        assert_equal(101, len(results))

    def insert_insert_get_indexed_slices(self):
        indexed_cf = ColumnFamily(pool, 'Indexed1')

        columns = {'birthdate': 1L}

        keys = []
        for i in range(1, 4):
            indexed_cf.insert('key%d' % i, columns)
            keys.append('key%d')

        expr = index.create_index_expression(column_name='birthdate', value=1L)
        clause = index.create_index_clause([expr])

        count = 0
        for key, cols in indexed_cf.get_indexed_slices(clause):
            assert_equal(cols, columns)
            assert key in keys
            count += 1
        assert_equal(count, 3)

    def test_get_indexed_slices_batching(self):
        indexed_cf = ColumnFamily(pool, 'Indexed1')

        columns = {'birthdate': 1L}

        for i in range(200):
            indexed_cf.insert('key%d' % i, columns)

        expr = index.create_index_expression(column_name='birthdate', value=1L)
        clause = index.create_index_clause([expr], count=10)

        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=2))
        assert_equal(len(result), 10)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=10))
        assert_equal(len(result), 10)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=77))
        assert_equal(len(result), 10)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=200))
        assert_equal(len(result), 10)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=1000))
        assert_equal(len(result), 10)

        clause = index.create_index_clause([expr], count=250)

        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=2))
        assert_equal(len(result), 200)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=10))
        assert_equal(len(result), 200)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=77))
        assert_equal(len(result), 200)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=200))
        assert_equal(len(result), 200)
        result = list(indexed_cf.get_indexed_slices(clause, buffer_size=1000))
        assert_equal(len(result), 200)

    def test_multiget_batching(self):
        key_prefix = "TestColumnFamily.test_multiget_batching"
        keys = []
        expected = OrderedDict()
        for i in range(10):
            key = key_prefix + str(i)
            keys.append(key)
            expected[key] = {'col': 'val'}
            cf.insert(key, {'col': 'val'})

        assert_equal(cf.multiget(keys, buffer_size=1), expected)
        assert_equal(cf.multiget(keys, buffer_size=2), expected)
        assert_equal(cf.multiget(keys, buffer_size=3), expected)
        assert_equal(cf.multiget(keys, buffer_size=9), expected)
        assert_equal(cf.multiget(keys, buffer_size=10), expected)
        assert_equal(cf.multiget(keys, buffer_size=11), expected)
        assert_equal(cf.multiget(keys, buffer_size=100), expected)

    def test_add(self):
        counter_cf.add('key', 'col')
        result = counter_cf.get('key')
        assert_equal(result['col'], 1)

        counter_cf.add('key', 'col')
        result = counter_cf.get('key')
        assert_equal(result['col'], 2)

        counter_cf.add('key', 'col2')
        result = counter_cf.get('key')
        assert_equal(result, {'col': 2, 'col2': 1})

    def test_insert_counters(self):
        counter_cf.insert('counter_key', {'col1': 1})
        result = counter_cf.get('counter_key')
        assert_equal(result['col1'], 1)

        counter_cf.insert('counter_key', {'col1': 1, 'col2': 1})
        result = counter_cf.get('counter_key')
        assert_equal(result, {'col1': 2, 'col2': 1})

    def test_remove(self):
        key = 'TestColumnFamily.test_remove'
        columns = {'1': 'val1', '2': 'val2'}
        cf.insert(key, columns)

        # An empty list for columns shouldn't delete anything
        cf.remove(key, columns=[])
        assert_equal(cf.get(key), columns)

        cf.remove(key, columns=['2'])
        del columns['2']
        assert_equal(cf.get(key), {'1': 'val1'})

        cf.remove(key)
        assert_raises(NotFoundException, cf.get, key)

    def test_remove_counter(self):
        key = 'test_remove_counter'
        counter_cf.add(key, 'col')
        result = counter_cf.get(key)
        assert_equal(result['col'], 1)

        counter_cf.remove_counter(key, 'col')
        assert_raises(NotFoundException, cf.get, key)

    def test_dict_class(self):
        key = 'TestColumnFamily.test_dict_class'
        cf.insert(key, {'1': 'val1'})
        assert isinstance(cf.get(key), TestDict)

    def test_xget(self):
        key = "test_xget_batching"
        cf.insert(key, dict((str(i), str(i)) for i in range(100, 300)))

        combos = [(100, 10),
                  (100, 1000),
                  (100, 199),
                  (100, 200),
                  (100, 201),
                  (100, 7),
                  (100, 2)]

        for count, bufsz in combos:
            res = list(cf.xget(key, column_count=count, buffer_size=bufsz))
            assert_equal(len(res), count)
            assert_equal(res, [(str(i), str(i)) for i in range(100, 200)])

        combos = [(10000, 2),
                  (10000, 7),
                  (10000, 199),
                  (10000, 200),
                  (10000, 201),
                  (10000, 10000)]

        for count, bufsz in combos:
            res = list(cf.xget(key, column_count=count, buffer_size=bufsz))
            assert_equal(len(res), 200)
            assert_equal(res, [(str(i), str(i)) for i in range(100, 300)])

        for bufsz in [2, 77, 199, 200, 201, 10000]:
            res = list(cf.xget(key, column_count=None, buffer_size=bufsz))
            assert_equal(len(res), 200)
            assert_equal(res, [(str(i), str(i)) for i in range(100, 300)])

    def test_xget_counter(self):
        key = 'test_xget_counter'
        counter_cf.insert(key, {'col1': 1})
        res = list(counter_cf.xget(key))
        assert_equal(res, [('col1', 1)])

        counter_cf.insert(key, {'col1': 1, 'col2': 1})
        res = list(counter_cf.xget(key))
        assert_equal(res, [('col1', 2), ('col2', 1)])

class TestSuperColumnFamily(unittest.TestCase):

    def tearDown(self):
        for key, columns in scf.get_range():
            scf.remove(key)

    def test_empty(self):
        key = 'TestSuperColumnFamily.test_empty'
        assert_raises(NotFoundException, cf.get, key)
        assert_equal(len(cf.multiget([key])), 0)
        for key, columns in cf.get_range():
            assert_equal(len(columns), 0)

    def test_get_whole_row(self):
        key = 'TestSuperColumnFamily.test_get_whole_row'
        columns = {'1': {'sub1': 'val1', 'sub2': 'val2'}, '2': {'sub3': 'val3', 'sub4': 'val4'}}
        scf.insert(key, columns)
        assert_equal(scf.get(key), columns)

    def test_get_super_column(self):
        key = 'TestSuperColumnFamily.test_get_super_column'
        subcolumns = {'sub1': 'val1', 'sub2': 'val2', 'sub3': 'val3'}
        columns = {'1': subcolumns}
        scf.insert(key, columns)
        assert_equal(scf.get(key), columns)
        assert_equal(scf.get(key, super_column='1'), subcolumns)
        assert_equal(scf.get(key, super_column='1', columns=['sub1']),     {'sub1': 'val1'})
        assert_equal(scf.get(key, super_column='1', column_start='sub3'),  {'sub3': 'val3'})
        assert_equal(scf.get(key, super_column='1', column_finish='sub1'), {'sub1': 'val1'})
        assert_equal(scf.get(key, super_column='1', column_count=1),       {'sub1': 'val1'})
        assert_equal(scf.get(key, super_column='1', column_count=1, column_reversed=True), {'sub3': 'val3'})

    def test_get_super_columns(self):
        key = 'TestSuperColumnFamily.test_get_super_columns'
        super1 = {'sub1': 'val1', 'sub2': 'val2'}
        super2 = {'sub3': 'val3', 'sub4': 'val4'}
        super3 = {'sub5': 'val5', 'sub6': 'val6'}
        columns = {'1': super1, '2': super2, '3': super3}
        scf.insert(key, columns)
        assert_equal(scf.get(key), columns)
        assert_equal(scf.get(key, columns=['1']),     {'1': super1})
        assert_equal(scf.get(key, column_start='3'),  {'3': super3})
        assert_equal(scf.get(key, column_finish='1'), {'1': super1})
        assert_equal(scf.get(key, column_count=1),    {'1': super1})
        assert_equal(scf.get(key, column_count=1, column_reversed=True), {'3': super3})

    def test_multiget_supercolumn(self):
        key1 = 'TestSuerColumnFamily.test_multiget_supercolumn1'
        key2 = 'TestSuerColumnFamily.test_multiget_supercolumn2'
        keys = [key1, key2]
        subcolumns = {'sub1': 'val1', 'sub2': 'val2', 'sub3': 'val3'}
        columns = {'1': subcolumns}
        scf.insert(key1, columns)
        scf.insert(key2, columns)

        assert_equal(scf.multiget(keys),
                     {key1: columns, key2: columns})

        assert_equal(scf.multiget(keys, super_column='1'),
                     {key1: subcolumns, key2: subcolumns})

        assert_equal(scf.multiget(keys, super_column='1', columns=['sub1']),
                     {key1: {'sub1': 'val1'}, key2: {'sub1': 'val1'}})

        assert_equal(scf.multiget(keys, super_column='1', column_start='sub3'),
                     {key1: {'sub3': 'val3'}, key2: {'sub3': 'val3'}})

        assert_equal(scf.multiget(keys, super_column='1', column_finish='sub1'),
                     {key1: {'sub1': 'val1'}, key2: {'sub1': 'val1'}})

        assert_equal(scf.multiget(keys, super_column='1', column_count=1),
                     {key1: {'sub1': 'val1'}, key2: {'sub1': 'val1'}})

        assert_equal(scf.multiget(keys, super_column='1', column_count=1, column_reversed=True),
                     {key1: {'sub3': 'val3'}, key2: {'sub3': 'val3'}})

    def test_multiget_supercolumns(self):
        key1 = 'TestSuerColumnFamily.test_multiget_supercolumns1'
        key2 = 'TestSuerColumnFamily.test_multiget_supercolumns2'
        keys = [key1, key2]
        super1 = {'sub1': 'val1', 'sub2': 'val2'}
        super2 = {'sub3': 'val3', 'sub4': 'val4'}
        super3 = {'sub5': 'val5', 'sub6': 'val6'}
        columns = {'1': super1, '2': super2, '3': super3}
        scf.insert(key1, columns)
        scf.insert(key2, columns)
        assert_equal(scf.multiget(keys), {key1: columns, key2: columns})
        assert_equal(scf.multiget(keys, columns=['1']),     {key1: {'1': super1}, key2: {'1': super1}})
        assert_equal(scf.multiget(keys, column_start='3'),  {key1: {'3': super3}, key2: {'3': super3}})
        assert_equal(scf.multiget(keys, column_finish='1'), {key1: {'1': super1}, key2: {'1': super1}})
        assert_equal(scf.multiget(keys, column_count=1),    {key1: {'1': super1}, key2: {'1': super1}})
        assert_equal(scf.multiget(keys, column_count=1, column_reversed=True), {key1: {'3': super3}, key2: {'3': super3}})

    def test_get_range_super_column(self):
        key = 'TestSuperColumnFamily.test_get_range_super_column'
        subcolumns = {'sub1': 'val1', 'sub2': 'val2', 'sub3': 'val3'}
        columns = {'1': subcolumns}
        scf.insert(key, columns)
        assert_equal(list(scf.get_range(start=key, finish=key, super_column='1')),
                     [(key, subcolumns)])
        assert_equal(list(scf.get_range(start=key, finish=key, super_column='1', columns=['sub1'])),
                     [(key, {'sub1': 'val1'})])
        assert_equal(list(scf.get_range(start=key, finish=key, super_column='1', column_start='sub3')),
                     [(key, {'sub3': 'val3'})])
        assert_equal(list(scf.get_range(start=key, finish=key, super_column='1', column_finish='sub1')),
                     [(key, {'sub1': 'val1'})])
        assert_equal(list(scf.get_range(start=key, finish=key, super_column='1', column_count=1)),
                     [(key, {'sub1': 'val1'})])
        assert_equal(list(scf.get_range(start=key, finish=key, super_column='1', column_count=1, column_reversed=True)),
                     [(key, {'sub3': 'val3'})])

    def test_get_range_super_columns(self):
        key = 'TestSuperColumnFamily.test_get_range_super_columns'
        super1 = {'sub1': 'val1', 'sub2': 'val2'}
        super2 = {'sub3': 'val3', 'sub4': 'val4'}
        super3 = {'sub5': 'val5', 'sub6': 'val6'}
        columns = {'1': super1, '2': super2, '3': super3}
        scf.insert(key, columns)
        assert_equal(list(scf.get_range(start=key, finish=key, columns=['1'])),
                     [(key, {'1': super1})])
        assert_equal(list(scf.get_range(start=key, finish=key, column_start='3')),
                     [(key, {'3': super3})])
        assert_equal(list(scf.get_range(start=key, finish=key, column_finish='1')),
                     [(key, {'1': super1})])
        assert_equal(list(scf.get_range(start=key, finish=key, column_count=1)),
                     [(key, {'1': super1})])
        assert_equal(list(scf.get_range(start=key, finish=key, column_count=1, column_reversed=True)),
                     [(key, {'3': super3})])

    def test_get_count_super_column(self):
        key = 'TestSuperColumnFamily.test_get_count_super_column'
        subcolumns = {'sub1': 'val1', 'sub2': 'val2', 'sub3': 'val3'}
        columns = {'1': subcolumns}
        scf.insert(key, columns)
        assert_equal(scf.get_count(key, super_column='1'),                       3)
        assert_equal(scf.get_count(key, super_column='1', columns=['sub1']),     1)
        assert_equal(scf.get_count(key, super_column='1', column_start='sub3'),  1)
        assert_equal(scf.get_count(key, super_column='1', column_finish='sub1'), 1)

    def test_get_count_super_columns(self):
        key = 'TestSuperColumnFamily.test_get_count_super_columns'
        columns = {'1': {'sub1': 'val1'}, '2': {'sub2': 'val2'}, '3': {'sub3': 'val3'}}
        scf.insert(key, columns)
        assert_equal(scf.get_count(key),                    3)
        assert_equal(scf.get_count(key, columns=['1']),     1)
        assert_equal(scf.get_count(key, column_start='3'),  1)
        assert_equal(scf.get_count(key, column_finish='1'), 1)

    def test_multiget_count_super_column(self):
        key1 = 'TestSuperColumnFamily.test_multiget_count_super_column1'
        key2 = 'TestSuperColumnFamily.test_multiget_count_super_column2'
        keys = [key1, key2]
        subcolumns = {'sub1': 'val1', 'sub2': 'val2', 'sub3': 'val3'}
        columns = {'1': subcolumns}
        scf.insert(key1, columns)
        scf.insert(key2, columns)
        assert_equal(scf.multiget_count(keys, super_column='1'),                       {key1: 3, key2: 3})
        assert_equal(scf.multiget_count(keys, super_column='1', columns=['sub1']),     {key1: 1, key2: 1})
        assert_equal(scf.multiget_count(keys, super_column='1', column_start='sub3'),  {key1: 1, key2: 1})
        assert_equal(scf.multiget_count(keys, super_column='1', column_finish='sub1'), {key1: 1, key2: 1})

    def test_multiget_count_super_columns(self):
        key1 = 'TestSuperColumnFamily.test_multiget_count_super_columns1'
        key2 = 'TestSuperColumnFamily.test_multiget_count_super_columns2'
        keys = [key1, key2]
        columns = {'1': {'sub1': 'val1'}, '2': {'sub2': 'val2'}, '3': {'sub3': 'val3'}}
        scf.insert(key1, columns)
        scf.insert(key2, columns)
        assert_equal(scf.multiget_count(keys),                    {key1: 3, key2: 3})
        assert_equal(scf.multiget_count(keys, columns=['1']),     {key1: 1, key2: 1})
        assert_equal(scf.multiget_count(keys, column_start='3'),  {key1: 1, key2: 1})
        assert_equal(scf.multiget_count(keys, column_finish='1'), {key1: 1, key2: 1})

    def test_batch_insert(self):
        key1 = 'TestSuperColumnFamily.test_batch_insert1'
        key2 = 'TestSuperColumnFamily.test_batch_insert2'
        columns = {'1': {'sub1': 'val1'}, '2': {'sub2': 'val2', 'sub3': 'val3'}}
        scf.batch_insert({key1: columns, key2: columns})
        assert_equal(scf.get(key1), columns)
        assert_equal(scf.get(key2), columns)

    def test_add(self):
        counter_scf.add('key', 'col', super_column='scol')
        result = counter_scf.get('key', super_column='scol')
        assert_equal(result['col'], 1)

        counter_scf.add('key', 'col', super_column='scol')
        result = counter_scf.get('key', super_column='scol')
        assert_equal(result['col'], 2)

        counter_scf.add('key', 'col2', super_column='scol')
        result = counter_scf.get('key', super_column='scol')
        assert_equal(result, {'col': 2, 'col2': 1})

    def test_remove(self):
        key = 'TestSuperColumnFamily.test_remove'
        columns = {'1': {'sub1': 'val1'}, '2': {'sub2': 'val2', 'sub3': 'val3'}, '3': {'sub4': 'val4'}}
        scf.insert(key, columns)
        assert_equal(scf.get_count(key), 3)
        scf.remove(key, super_column='1')
        assert_equal(scf.get_count(key), 2)
        scf.remove(key, columns=['3'])
        assert_equal(scf.get_count(key), 1)

        assert_equal(scf.get_count(key, super_column='2'), 2)
        scf.remove(key, super_column='2', columns=['sub2'])
        assert_equal(scf.get_count(key, super_column='2'), 1)

    def test_remove_counter(self):
        key = 'test_remove_counter'
        counter_scf.add(key, 'col', super_column='scol')
        result = counter_scf.get(key, super_column='scol')
        assert_equal(result['col'], 1)

        counter_scf.remove_counter(key, 'col', super_column='scol')
        assert_raises(NotFoundException, scf.get, key)

    def test_xget_counter(self):
        key = 'test_xget_counter'
        counter_scf.insert(key, {'scol': {'col1': 1}})
        res = list(counter_scf.xget(key))
        assert_equal(res, [('scol', {'col1': 1})])

        counter_scf.insert(key, {'scol': {'col1': 1, 'col2': 1}})
        res = list(counter_scf.xget(key))
        assert_equal(res, [('scol', {'col1': 2, 'col2': 1})])



from datetime import datetime
import unittest
import uuid

from nose.tools import assert_raises, assert_equal, assert_true
from nose.plugins.skip import SkipTest

import pycassa.types as types
from pycassa import index, ColumnFamily, ConnectionPool, \
    ColumnFamilyMap, NotFoundException, SystemManager

from tests.util import requireOPP

CF = 'Standard1'
SCF = 'Super1'
INDEXED_CF = 'Indexed1'
pool = None
sys_man = None

def setup_module():
    global pool, sys_man
    credentials = {'username': 'jsmith', 'password': 'havebadpass'}
    pool = ConnectionPool(keyspace='PycassaTestKeyspace',
            credentials=credentials, timeout=1.0)
    sys_man = SystemManager()

def teardown_module():
    pool.dispose()


class TestUTF8(object):
    key = types.LexicalUUIDType()
    strcol = types.AsciiType(default='default')
    intcol = types.LongType(default=0)
    floatcol = types.FloatType(default=0.0)
    datetimecol = types.DateType()

    def __str__(self):
        return str(map(str, [self.strcol, self.intcol, self.floatcol, self.datetimecol]))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__


class TestIndex(object):
    birthdate = types.LongType(default=0)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__


class TestEmpty(object):
    pass


class TestColumnFamilyMap(unittest.TestCase):

    def setUp(self):
        self.sys_man = sys_man
        self.map = ColumnFamilyMap(TestUTF8, pool, CF)
        self.indexed_map = ColumnFamilyMap(TestIndex, pool, INDEXED_CF)
        self.empty_map = ColumnFamilyMap(TestEmpty, pool, CF, raw_columns=True)

    def tearDown(self):
        for instance in self.map.get_range():
            self.map.remove(instance)
        for instance in self.indexed_map.get_range():
            self.indexed_map.remove(instance)

    def instance(self):
        instance = TestUTF8()
        instance.key = uuid.uuid4()
        instance.strcol = '1'
        instance.intcol = 2
        instance.floatcol = 3.5
        instance.datetimecol = datetime.now().replace(microsecond=0)

        return instance

    def test_empty(self):
        key = uuid.uuid4()
        assert_raises(NotFoundException, self.map.get, key)
        assert_equal(len(self.map.multiget([key])), 0)

    def test_insert_get(self):
        instance = self.instance()
        assert_raises(NotFoundException, self.map.get, instance.key)
        ts = self.map.insert(instance)
        assert_true(isinstance(ts, (int, long)))
        assert_equal(self.map.get(instance.key), instance)

    def test_insert_get_omitting_columns(self):
        """
        When omitting columns, pycassa should not try to insert the CassandraType
        instance on a ColumnFamilyMap object
        """
        instance2 = TestUTF8()
        instance2.key = uuid.uuid4()
        instance2.strcol = 'lol'
        instance2.intcol = 2
        assert_raises(NotFoundException, self.map.get, instance2.key)
        self.map.insert(instance2)
        ret_inst = self.map.get(instance2.key)
        assert_equal(ret_inst.key, instance2.key)
        assert_equal(ret_inst.strcol, instance2.strcol)
        assert_equal(ret_inst.intcol, instance2.intcol)

        ## these lines are commented out because, though they should work, wont
        ## because CassandraTypes are not descriptors when used on a ColumnFamilyMap
        ## instance, they are merely class attributes that are overwritten at runtime

        # assert_equal(ret_inst.floatcol, instance2.floatcol)
        # assert_equal(ret_inst.datetimecol, instance2.datetimecol)
        # assert_equal(self.map.get(instance2.key), instance2)

    def test_insert_get_indexed_slices(self):
        instance1 = TestIndex()
        instance1.key = 'key1'
        instance1.birthdate = 1L
        self.indexed_map.insert(instance1)

        instance2 = TestIndex()
        instance2.key = 'key2'
        instance2.birthdate = 1L
        self.indexed_map.insert(instance2)

        instance3 = TestIndex()
        instance3.key = 'key3'
        instance3.birthdate = 2L
        self.indexed_map.insert(instance3)

        expr = index.create_index_expression(column_name='birthdate', value=2L)
        clause = index.create_index_clause([expr])

        result = self.indexed_map.get_indexed_slices(index_clause=clause)
        count = 0
        for instance in result:
            assert_equal(instance, instance3)
            count += 1
        assert_equal(count, 1)

    def test_insert_multiget(self):
        instance1 = self.instance()
        instance2 = self.instance()
        missing_key = uuid.uuid4()

        self.map.insert(instance1)
        self.map.insert(instance2)
        rows = self.map.multiget([instance1.key, instance2.key, missing_key])
        assert_equal(len(rows), 2)
        assert_equal(rows[instance1.key], instance1)
        assert_equal(rows[instance2.key], instance2)
        assert_true(missing_key not in rows)

    @requireOPP
    def test_insert_get_range(self):
        instances = [self.instance() for i in range(5)]
        instances = sorted(instances, key=lambda instance: instance.key)
        for instance in instances:
            self.map.insert(instance)

        rows = list(self.map.get_range(start=instances[0].key, finish=instances[-1].key))
        assert_equal(len(rows), len(instances))
        assert_equal(rows, instances)

    def test_remove(self):
        instance = self.instance()

        self.map.insert(instance)
        self.map.remove(instance)
        assert_raises(NotFoundException, self.map.get, instance.key)

    def test_does_not_insert_extra_column(self):
        instance = self.instance()
        instance.othercol = 'Test'

        self.map.insert(instance)

        get_instance = self.map.get(instance.key)
        assert_equal(get_instance.strcol, instance.strcol)
        assert_equal(get_instance.intcol, instance.intcol)
        assert_equal(get_instance.floatcol, instance.floatcol)
        assert_equal(get_instance.datetimecol, instance.datetimecol)
        assert_raises(AttributeError, getattr, get_instance, 'othercol')

    def test_has_defaults(self):
        key = uuid.uuid4()
        ColumnFamily.insert(self.map, key, {'strcol': '1'})
        instance = self.map.get(key)

        assert_equal(instance.intcol, TestUTF8.intcol.default)
        assert_equal(instance.floatcol, TestUTF8.floatcol.default)
        assert_equal(instance.datetimecol, TestUTF8.datetimecol.default)

    def test_batch_insert(self):
        instances = []
        for i in range(3):
            instance = TestUTF8()
            instance.key = uuid.uuid4()
            instance.strcol = 'instance%s' % (i + 1)
            instances.append(instance)

        for i in instances:
            assert_raises(NotFoundException, self.map.get, i.key)

        self.map.batch_insert(instances)

        for i in instances:
            get_instance = self.map.get(i.key)
            assert_equal(get_instance.key, i.key)
            assert_equal(get_instance.strcol, i.strcol)

class TestSuperColumnFamilyMap(unittest.TestCase):

    def setUp(self):
        self.map = ColumnFamilyMap(TestUTF8, pool, SCF)

    def tearDown(self):
        for scols in self.map.get_range():
            for instance in scols.values():
                self.map.remove(instance)

    def instance(self, super_column):
        instance = TestUTF8()
        instance.key = uuid.uuid4()
        instance.super_column = super_column
        instance.strcol = '1'
        instance.intcol = 2
        instance.floatcol = 3.5
        instance.datetimecol = datetime.now().replace(microsecond=0)

        return instance

    def test_super(self):
        instance = self.instance('super1')
        assert_raises(NotFoundException, self.map.get, instance.key)
        self.map.insert(instance)
        res = self.map.get(instance.key)[instance.super_column]
        assert_equal(res, instance)
        assert_equal(self.map.multiget([instance.key])[instance.key][instance.super_column], instance)
        assert_equal(list(self.map.get_range(start=instance.key, finish=instance.key)), [{instance.super_column: instance}])

    def test_super_remove(self):
        instance1 = self.instance('super1')
        assert_raises(NotFoundException, self.map.get, instance1.key)
        self.map.insert(instance1)

        instance2 = self.instance('super2')
        self.map.insert(instance2)

        self.map.remove(instance2)
        assert_equal(len(self.map.get(instance1.key)), 1)
        assert_equal(self.map.get(instance1.key)[instance1.super_column], instance1)

    def test_batch_insert_super(self):
        instances = []
        for i in range(3):
            instance = self.instance('super_batch%s' % (i + 1))
            instances.append(instance)

        for i in instances:
            assert_raises(NotFoundException, self.map.get, i.key)

        self.map.batch_insert(instances)

        for i in instances:
            result = self.map.get(i.key)
            get_instance = result[i.super_column]
            assert_equal(len(result), 1)
            assert_equal(get_instance.key, i.key)
            assert_equal(get_instance.super_column, i.super_column)
            assert_equal(get_instance.strcol, i.strcol)



from pycassa.system_manager import SystemManager

TEST_KS = 'PycassaTestKeyspace'

def setup_package():
    sys = SystemManager()
    if TEST_KS in sys.list_keyspaces():
        sys.drop_keyspace(TEST_KS)
    try:
        sys.create_keyspace(TEST_KS, 'SimpleStrategy', {'replication_factor': '1'})
        sys.create_column_family(TEST_KS, 'Standard1')
        sys.create_column_family(TEST_KS, 'Super1', super=True)
        sys.create_column_family(TEST_KS, 'Indexed1')
        sys.create_index(TEST_KS, 'Indexed1', 'birthdate', 'LongType')
        sys.create_column_family(TEST_KS, 'Counter1',
                                 default_validation_class='CounterColumnType')
        sys.create_column_family(TEST_KS, 'SuperCounter1', super=True,
                                 default_validation_class='CounterColumnType')
    except Exception, e:
        print e
        try:
            sys.drop_keyspace(TEST_KS)
        except:
            pass
        raise e
    sys.close()

def teardown_package():
    sys = SystemManager()
    if TEST_KS in sys.list_keyspaces():
        sys.drop_keyspace(TEST_KS)
    sys.close()



import threading
import unittest
import time

from nose.tools import assert_raises, assert_equal, assert_true
from pycassa import ColumnFamily, ConnectionPool, InvalidRequestError,\
                    NoConnectionAvailable, MaximumRetryException, AllServersUnavailable
from pycassa.logging.pool_stats_logger import StatsLogger
from pycassa.cassandra.ttypes import ColumnPath
from pycassa.cassandra.ttypes import InvalidRequestException
from pycassa.cassandra.ttypes import NotFoundException


_credentials = {'username': 'jsmith', 'password': 'havebadpass'}

def _get_list():
    return ['foo:bar']

class PoolingCase(unittest.TestCase):

    def tearDown(self):
        pool = ConnectionPool('PycassaTestKeyspace')
        cf = ColumnFamily(pool, 'Standard1')
        for key, cols in cf.get_range():
            cf.remove(key)

    def test_basic_pools(self):
        pool = ConnectionPool('PycassaTestKeyspace', credentials=_credentials)
        cf = ColumnFamily(pool, 'Standard1')
        cf.insert('key1', {'col': 'val'})
        pool.dispose()

    def test_empty_list(self):
        assert_raises(AllServersUnavailable, ConnectionPool, 'PycassaTestKeyspace', server_list=[])

    def test_server_list_func(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool('PycassaTestKeyspace', server_list=_get_list,
                         listeners=[stats_logger], prefill=False)
        assert_equal(stats_logger.serv_list, ['foo:bar'])
        assert_equal(stats_logger.stats['list'], 1)
        pool.dispose()

    def test_queue_pool(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=True, pool_timeout=0.1, timeout=1,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=False)
        conns = []
        for i in range(10):
            conns.append(pool.get())

        assert_equal(stats_logger.stats['created']['success'], 10)
        assert_equal(stats_logger.stats['checked_out'], 10)

        # Pool is maxed out now
        assert_raises(NoConnectionAvailable, pool.get)
        assert_equal(stats_logger.stats['created']['success'], 10)
        assert_equal(stats_logger.stats['at_max'], 1)

        for i in range(0, 5):
            pool.return_conn(conns[i])
        assert_equal(stats_logger.stats['disposed']['success'], 0)
        assert_equal(stats_logger.stats['checked_in'], 5)

        for i in range(5, 10):
            pool.return_conn(conns[i])
        assert_equal(stats_logger.stats['disposed']['success'], 5)
        assert_equal(stats_logger.stats['checked_in'], 10)

        conns = []

        # These connections should come from the pool
        for i in range(5):
            conns.append(pool.get())
        assert_equal(stats_logger.stats['created']['success'], 10)
        assert_equal(stats_logger.stats['checked_out'], 15)

        # But these will need to be made
        for i in range(5):
            conns.append(pool.get())
        assert_equal(stats_logger.stats['created']['success'], 15)
        assert_equal(stats_logger.stats['checked_out'], 20)

        assert_equal(stats_logger.stats['disposed']['success'], 5)
        for i in range(10):
            conns[i].return_to_pool()
        assert_equal(stats_logger.stats['checked_in'], 20)
        assert_equal(stats_logger.stats['disposed']['success'], 10)

        assert_raises(InvalidRequestError, conns[0].return_to_pool)
        assert_equal(stats_logger.stats['checked_in'], 20)
        assert_equal(stats_logger.stats['disposed']['success'], 10)

        print "in test:", id(conns[-1])
        conns[-1].return_to_pool()
        assert_equal(stats_logger.stats['checked_in'], 20)
        assert_equal(stats_logger.stats['disposed']['success'], 10)

        pool.dispose()

    def test_queue_pool_threadlocal(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=True, pool_timeout=0.01, timeout=1,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=True)
        conns = []

        assert_equal(stats_logger.stats['created']['success'], 5)
        # These connections should all be the same
        for i in range(10):
            conns.append(pool.get())
        assert_equal(stats_logger.stats['created']['success'], 5)
        assert_equal(stats_logger.stats['checked_out'], 1)

        for i in range(0, 5):
            pool.return_conn(conns[i])
        assert_equal(stats_logger.stats['checked_in'], 1)
        for i in range(5, 10):
            pool.return_conn(conns[i])
        assert_equal(stats_logger.stats['checked_in'], 1)

        conns = []

        assert_equal(stats_logger.stats['created']['success'], 5)
        # A single connection should come from the pool
        for i in range(5):
            conns.append(pool.get())
        assert_equal(stats_logger.stats['created']['success'], 5)
        assert_equal(stats_logger.stats['checked_out'], 2)

        for conn in conns:
            pool.return_conn(conn)

        conns = []
        threads = []
        stats_logger.reset()

        def checkout_return():
            conn = pool.get()
            time.sleep(1)
            pool.return_conn(conn)

        for i in range(5):
            threads.append(threading.Thread(target=checkout_return))
            threads[-1].start()
        for thread in threads:
            thread.join()

        assert_equal(stats_logger.stats['created']['success'], 0) # Still 5 connections in pool
        assert_equal(stats_logger.stats['checked_out'], 5)
        assert_equal(stats_logger.stats['checked_in'], 5)

        # These should come from the pool
        threads = []
        for i in range(5):
            threads.append(threading.Thread(target=checkout_return))
            threads[-1].start()
        for thread in threads:
            thread.join()
        assert_equal(stats_logger.stats['created']['success'], 0)
        assert_equal(stats_logger.stats['checked_out'], 10)
        assert_equal(stats_logger.stats['checked_in'], 10)

        pool.dispose()

    def test_queue_pool_no_prefill(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=False, pool_timeout=0.1, timeout=1,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=False)
        conns = []
        for i in range(10):
            conns.append(pool.get())
            assert_equal(stats_logger.stats['created']['success'], i + 1)
            assert_equal(stats_logger.stats['checked_out'], i + 1)

        # Pool is maxed out now
        assert_raises(NoConnectionAvailable, pool.get)
        assert_equal(stats_logger.stats['created']['success'], 10)
        assert_equal(stats_logger.stats['at_max'], 1)

        for i in range(0, 5):
            pool.return_conn(conns[i])
            assert_equal(stats_logger.stats['checked_in'], i + 1)
            assert_equal(stats_logger.stats['disposed']['success'], 0)

        for i in range(5, 10):
            pool.return_conn(conns[i])
            assert_equal(stats_logger.stats['checked_in'], i + 1)
            assert_equal(stats_logger.stats['disposed']['success'], (i - 5) + 1)

        conns = []

        # These connections should come from the pool
        for i in range(5):
            conns.append(pool.get())
            assert_equal(stats_logger.stats['created']['success'], 10)
            assert_equal(stats_logger.stats['checked_out'], (i + 10) + 1)

        # But these will need to be made
        for i in range(5):
            conns.append(pool.get())
            assert_equal(stats_logger.stats['created']['success'], (i + 10) + 1)
            assert_equal(stats_logger.stats['checked_out'], (i + 15) + 1)

        assert_equal(stats_logger.stats['disposed']['success'], 5)
        for i in range(10):
            conns[i].return_to_pool()
            assert_equal(stats_logger.stats['checked_in'], (i + 10) + 1)
        assert_equal(stats_logger.stats['disposed']['success'], 10)

        # Make sure a double return doesn't change our counts
        assert_raises(InvalidRequestError, conns[0].return_to_pool)
        assert_equal(stats_logger.stats['checked_in'], 20)
        assert_equal(stats_logger.stats['disposed']['success'], 10)

        conns[-1].return_to_pool()
        assert_equal(stats_logger.stats['checked_in'], 20)
        assert_equal(stats_logger.stats['disposed']['success'], 10)

        pool.dispose()

    def test_queue_pool_recycle(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=1,
                         prefill=True, pool_timeout=0.5, timeout=1,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=False)

        cf = ColumnFamily(pool, 'Standard1')
        columns = {'col1': 'val', 'col2': 'val'}
        for i in range(10):
            cf.insert('key', columns)

        assert_equal(stats_logger.stats['recycled'], 5)

        pool.dispose()
        stats_logger.reset()

        # Try with threadlocal=True
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=1,
                         prefill=False, pool_timeout=0.5, timeout=1,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=True)

        cf = ColumnFamily(pool, 'Standard1')
        for i in range(10):
            cf.insert('key', columns)

        pool.dispose()
        assert_equal(stats_logger.stats['recycled'], 5)

    def test_pool_connection_failure(self):
        stats_logger = StatsLoggerWithListStorage()

        def get_extra():
            """Make failure count adjustments based on whether or not
            the permuted list starts with a good host:port"""
            if stats_logger.serv_list[0] == 'localhost:9160':
                return 0
            else:
                return 1

        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000, prefill=True,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         pool_timeout=0.01, timeout=0.05,
                         listeners=[stats_logger], use_threadlocal=False,
                         server_list=['localhost:9160', 'foobar:1'])

        assert_equal(stats_logger.stats['failed'], 4 + get_extra())

        for i in range(0, 7):
            pool.get()

        assert_equal(stats_logger.stats['failed'], 6 + get_extra())

        pool.dispose()
        stats_logger.reset()

        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000, prefill=True,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         pool_timeout=0.01, timeout=0.05,
                         listeners=[stats_logger], use_threadlocal=True,
                         server_list=['localhost:9160', 'foobar:1'])

        assert_equal(stats_logger.stats['failed'], 4 + get_extra())

        threads = []
        for i in range(0, 7):
            threads.append(threading.Thread(target=pool.get))
            threads[-1].start()
        for thread in threads:
            thread.join()

        assert_equal(stats_logger.stats['failed'], 6 + get_extra())

        pool.dispose()

    def test_queue_failover(self):
        for prefill in (True, False):
            stats_logger = StatsLoggerWithListStorage()
            pool = ConnectionPool(pool_size=1, max_overflow=0, recycle=10000,
                             prefill=prefill, timeout=1,
                             keyspace='PycassaTestKeyspace', credentials=_credentials,
                             listeners=[stats_logger], use_threadlocal=False,
                             server_list=['localhost:9160', 'localhost:9160'])

            cf = ColumnFamily(pool, 'Standard1')

            for i in range(1, 5):
                conn = pool.get()
                setattr(conn, 'send_batch_mutate', conn._fail_once)
                conn._should_fail = True
                conn.return_to_pool()

                # The first insert attempt should fail, but failover should occur
                # and the insert should succeed
                cf.insert('key', {'col': 'val%d' % i, 'col2': 'val'})
                assert_equal(stats_logger.stats['failed'], i)
                assert_equal(cf.get('key'), {'col': 'val%d' % i, 'col2': 'val'})

            pool.dispose()

    def test_queue_threadlocal_failover(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=1, max_overflow=0, recycle=10000,
                         prefill=True, timeout=0.05,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=True,
                         server_list=['localhost:9160', 'localhost:9160'])

        cf = ColumnFamily(pool, 'Standard1')

        for i in range(1, 5):
            conn = pool.get()
            setattr(conn, 'send_batch_mutate', conn._fail_once)
            conn._should_fail = True
            conn.return_to_pool()

            # The first insert attempt should fail, but failover should occur
            # and the insert should succeed
            cf.insert('key', {'col': 'val%d' % i, 'col2': 'val'})
            assert_equal(stats_logger.stats['failed'], i)
            assert_equal(cf.get('key'), {'col': 'val%d' % i, 'col2': 'val'})

        pool.dispose()
        stats_logger.reset()

        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=True, timeout=0.05,
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=True,
                         server_list=['localhost:9160', 'localhost:9160'])

        cf = ColumnFamily(pool, 'Standard1')

        for i in range(5):
            conn = pool.get()
            setattr(conn, 'send_batch_mutate', conn._fail_once)
            conn._should_fail = True
            conn.return_to_pool()

        threads = []
        args = ('key', {'col': 'val', 'col2': 'val'})
        for i in range(5):
            threads.append(threading.Thread(target=cf.insert, args=args))
            threads[-1].start()
        for thread in threads:
            thread.join()

        assert_equal(stats_logger.stats['failed'], 5)

        pool.dispose()

    def test_queue_retry_limit(self):
        for prefill in (True, False):
            stats_logger = StatsLoggerWithListStorage()
            pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                             prefill=prefill, max_retries=3, # allow 3 retries
                             keyspace='PycassaTestKeyspace', credentials=_credentials,
                             listeners=[stats_logger], use_threadlocal=False,
                             server_list=['localhost:9160', 'localhost:9160'])

            # Corrupt all of the connections
            for i in range(5):
                conn = pool.get()
                setattr(conn, 'send_batch_mutate', conn._fail_once)
                conn._should_fail = True
                conn.return_to_pool()

            cf = ColumnFamily(pool, 'Standard1')
            assert_raises(MaximumRetryException, cf.insert, 'key', {'col': 'val', 'col2': 'val'})
            assert_equal(stats_logger.stats['failed'], 4) # On the 4th failure, didn't retry

            pool.dispose()

    def test_queue_failure_on_retry(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=True, max_retries=3, # allow 3 retries
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=False,
                         server_list=['localhost:9160', 'localhost:9160'])

        def raiser():
            raise IOError
        # Replace wrapper will open a connection to get the version, so if it
        # fails we need to retry as with any other connection failure
        pool._replace_wrapper = raiser

        # Corrupt all of the connections
        for i in range(5):
            conn = pool.get()
            setattr(conn, 'send_batch_mutate', conn._fail_once)
            conn._should_fail = True
            conn.return_to_pool()

        cf = ColumnFamily(pool, 'Standard1')
        assert_raises(MaximumRetryException, cf.insert, 'key', {'col': 'val', 'col2': 'val'})
        assert_equal(stats_logger.stats['failed'], 4) # On the 4th failure, didn't retry

        pool.dispose()

    def test_queue_threadlocal_retry_limit(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=True, max_retries=3, # allow 3 retries
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=True,
                         server_list=['localhost:9160', 'localhost:9160'])

        # Corrupt all of the connections
        for i in range(5):
            conn = pool.get()
            setattr(conn, 'send_batch_mutate', conn._fail_once)
            conn._should_fail = True
            conn.return_to_pool()

        cf = ColumnFamily(pool, 'Standard1')
        assert_raises(MaximumRetryException, cf.insert, 'key', {'col': 'val', 'col2': 'val'})
        assert_equal(stats_logger.stats['failed'], 4) # On the 4th failure, didn't retry

        pool.dispose()

    def test_queue_failure_with_no_retries(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                         prefill=True, max_retries=3, # allow 3 retries
                         keyspace='PycassaTestKeyspace', credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=False,
                         server_list=['localhost:9160', 'localhost:9160'])

        # Corrupt all of the connections
        for i in range(5):
            conn = pool.get()
            setattr(conn, 'send_batch_mutate', conn._fail_once)
            conn._should_fail = True
            conn.return_to_pool()

        cf = ColumnFamily(pool, 'Counter1')
        assert_raises(MaximumRetryException, cf.insert, 'key', {'col': 2, 'col2': 2})
        assert_equal(stats_logger.stats['failed'], 1)  # didn't retry at all

        pool.dispose()

    def test_failure_connection_info(self):
        stats_logger = StatsLoggerRequestInfo()
        pool = ConnectionPool(pool_size=1, max_overflow=0, recycle=10000,
                              prefill=True, max_retries=0,
                              keyspace='PycassaTestKeyspace', credentials=_credentials,
                              listeners=[stats_logger], use_threadlocal=True,
                              server_list=['localhost:9160'])
        cf = ColumnFamily(pool, 'Counter1')

        # Corrupt the connection
        conn = pool.get()
        setattr(conn, 'send_get', conn._fail_once)
        conn._should_fail = True
        conn.return_to_pool()

        assert_raises(MaximumRetryException, cf.get, 'greunt', columns=['col'])
        assert_true('request' in stats_logger.failure_dict['connection'].info)
        request = stats_logger.failure_dict['connection'].info['request']
        assert_equal(request['method'], 'get')
        assert_equal(request['args'], ('greunt', ColumnPath('Counter1', None, 'col'), 1))
        assert_equal(request['kwargs'], {})

    def test_pool_invalid_request(self):
        stats_logger = StatsLoggerWithListStorage()
        pool = ConnectionPool(pool_size=1, max_overflow=0, recycle=10000,
                         prefill=True, max_retries=3,
                         keyspace='PycassaTestKeyspace',
                         credentials=_credentials,
                         listeners=[stats_logger], use_threadlocal=False,
                         server_list=['localhost:9160'])
        cf = ColumnFamily(pool, 'Standard1')
        # Make sure the pool doesn't hide and retries invalid requests
        assert_raises(InvalidRequestException, cf.add, 'key', 'col')
        assert_raises(NotFoundException, cf.get, 'none')
        pool.dispose()


class StatsLoggerWithListStorage(StatsLogger):

    def obtained_server_list(self, dic):
        StatsLogger.obtained_server_list(self, dic)
        self.serv_list = dic.get('server_list')


class StatsLoggerRequestInfo(StatsLogger):

    def connection_failed(self, dic):
        StatsLogger.connection_failed(self, dic)
        self.failure_dict = dic



from unittest import TestCase
from nose.tools import assert_equal, assert_raises

from pycassa.logging.pool_stats_logger import StatsLogger
from pycassa.pool import ConnectionPool, NoConnectionAvailable, InvalidRequestError

__author__ = 'gilles'

_credentials = {'username': 'jsmith', 'password': 'havebadpass'}

class TestStatsLogger(TestCase):
    def __init__(self, methodName='runTest'):
        super(TestStatsLogger, self).__init__(methodName)

    def setUp(self):
        super(TestStatsLogger, self).setUp()
        self.logger = StatsLogger()

    def test_empty(self):
        assert_equal(self.logger.stats, self.logger._stats)

    def test_connection_created(self):
        self.logger.connection_created({'level': 'info'})
        self.logger.connection_created({'level': 'error'})

        stats = self.logger.stats
        assert_equal(stats['created']['success'], 1)
        assert_equal(stats['created']['failure'], 1)

    def test_connection_checked(self):
        self.logger.connection_checked_out({})
        self.logger.connection_checked_out({})
        self.logger.connection_checked_in({})
        stats = self.logger.stats
        assert_equal(stats['checked_out'], 2)
        assert_equal(stats['checked_in'], 1)
        assert_equal(stats['opened'], {'current': 1, 'max': 2})

    def test_connection_disposed(self):
        self.logger.connection_disposed({'level': 'info'})
        self.logger.connection_disposed({'level': 'error'})

        stats = self.logger.stats
        assert_equal(stats['disposed']['success'], 1)
        assert_equal(stats['disposed']['failure'], 1)

    def test_connection_recycled(self):
        self.logger.connection_recycled({})
        stats = self.logger.stats
        assert_equal(stats['recycled'], 1)

    def test_connection_failed(self):
        self.logger.connection_failed({})
        stats = self.logger.stats
        assert_equal(stats['failed'], 1)

    def test_obtained_server_list(self):
        self.logger.obtained_server_list({})
        stats = self.logger.stats
        assert_equal(stats['list'], 1)

    def test_pool_at_max(self):
        self.logger.pool_at_max({})
        stats = self.logger.stats
        assert_equal(stats['at_max'], 1)


class TestInPool(TestCase):
    def __init__(self, methodName='runTest'):
        super(TestInPool, self).__init__(methodName)

    def test_pool(self):
        listener = StatsLogger()
        pool = ConnectionPool(pool_size=5, max_overflow=5, recycle=10000,
                              prefill=True, pool_timeout=0.1, timeout=1,
                              keyspace='PycassaTestKeyspace', credentials=_credentials,
                              listeners=[listener], use_threadlocal=False)
        conns = []
        for i in range(10):
            conns.append(pool.get())
        assert_equal(listener.stats['created']['success'], 10)
        assert_equal(listener.stats['created']['failure'], 0)
        assert_equal(listener.stats['checked_out'], 10)
        assert_equal(listener.stats['opened'], {'current': 10, 'max': 10})

        # Pool is maxed out now
        assert_raises(NoConnectionAvailable, pool.get)
        assert_equal(listener.stats['created']['success'], 10)
        assert_equal(listener.stats['checked_out'], 10)
        assert_equal(listener.stats['opened'], {'current': 10, 'max': 10})
        assert_equal(listener.stats['at_max'], 1)

        for i in range(0, 5):
            pool.return_conn(conns[i])
        assert_equal(listener.stats['disposed']['success'], 0)
        assert_equal(listener.stats['checked_in'], 5)
        assert_equal(listener.stats['opened'], {'current': 5, 'max': 10})

        for i in range(5, 10):
            pool.return_conn(conns[i])
        assert_equal(listener.stats['disposed']['success'], 5)
        assert_equal(listener.stats['checked_in'], 10)

        conns = []

        # These connections should come from the pool
        for i in range(5):
            conns.append(pool.get())
        assert_equal(listener.stats['created']['success'], 10)
        assert_equal(listener.stats['checked_out'], 15)

        # But these will need to be made
        for i in range(5):
            conns.append(pool.get())
        assert_equal(listener.stats['created']['success'], 15)
        assert_equal(listener.stats['checked_out'], 20)

        assert_equal(listener.stats['disposed']['success'], 5)
        for i in range(10):
            conns[i].return_to_pool()
        assert_equal(listener.stats['checked_in'], 20)
        assert_equal(listener.stats['disposed']['success'], 10)

        assert_raises(InvalidRequestError, conns[0].return_to_pool)
        assert_equal(listener.stats['checked_in'], 20)
        assert_equal(listener.stats['disposed']['success'], 10)

        print "in test:", id(conns[-1])
        conns[-1].return_to_pool()
        assert_equal(listener.stats['checked_in'], 20)
        assert_equal(listener.stats['disposed']['success'], 10)

        pool.dispose()



import unittest
import time

from nose.tools import assert_raises, assert_equal, assert_true

from pycassa import index, ColumnFamily, ConnectionPool,\
                    NotFoundException
from pycassa.contrib.stubs import ColumnFamilyStub, ConnectionPoolStub
from pycassa.util import convert_time_to_uuid

pool = cf = indexed_cf = None
pool_stub = cf_stub = indexed_cf_stub = None


def setup_module():
    global pool, cf, indexed_cf, pool_stub, indexed_cf_stub, cf_stub
    credentials = {'username': 'jsmith', 'password': 'havebadpass'}
    pool = ConnectionPool(keyspace='PycassaTestKeyspace',
            credentials=credentials, timeout=1.0)
    cf = ColumnFamily(pool, 'Standard1', dict_class=TestDict)
    indexed_cf = ColumnFamily(pool, 'Indexed1')

    pool_stub = ConnectionPoolStub(keyspace='PycassaTestKeyspace',
            credentials=credentials, timeout=1.0)
    cf_stub = ColumnFamilyStub(pool_stub, 'Standard1', dict_class=TestDict)
    indexed_cf_stub = ColumnFamilyStub(pool_stub, 'Indexed1')


def teardown_module():
    cf.truncate()
    cf_stub.truncate()
    indexed_cf.truncate()
    indexed_cf_stub.truncate()
    pool.dispose()


class TestDict(dict):
    pass


class TestColumnFamilyStub(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        for test_cf in (cf, cf_stub):
            for key, columns in test_cf.get_range():
                test_cf.remove(key)

    def test_empty(self):
        key = 'TestColumnFamily.test_empty'

        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            assert_equal(len(test_cf.multiget([key])), 0)
            for key, columns in test_cf.get_range():
                assert_equal(len(columns), 0)

    def test_insert_get(self):
        key = 'TestColumnFamily.test_insert_get'
        columns = {'1': 'val1', '2': 'val2'}
        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            ts = test_cf.insert(key, columns)
            assert_true(isinstance(ts, (int, long)))
            assert_equal(test_cf.get(key), columns)

    def test_insert_get_column_start_and_finish_reversed(self):
        key = 'TestColumnFamily.test_insert_get_reversed'
        columns = {'1': 'val1', '2': 'val2'}
        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            ts = test_cf.insert(key, columns)
            assert_true(isinstance(ts, (int, long)))
            test_cf.get(key, column_reversed=True)

    def test_insert_get_column_start_and_finish(self):
        key = 'TestColumnFamily.test_insert_get_column_start_and_finish'
        columns = {'a': 'val1', 'b': 'val2', 'c': 'val3', 'd': 'val4'}
        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            ts = test_cf.insert(key, columns)
            assert_true(isinstance(ts, (int, long)))
            assert_equal(test_cf.get(key, column_start='b', column_finish='c'), {'b': 'val2', 'c': 'val3'})

    def test_insert_get_column_start_and_reversed(self):
        key = 'TestColumnFamily.test_insert_get_column_start_and_finish_reversed'
        columns = {'a': 'val1', 'b': 'val2', 'c': 'val3', 'd': 'val4'}
        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            ts = test_cf.insert(key, columns)
            assert_true(isinstance(ts, (int, long)))
            assert_equal(test_cf.get(key, column_start='b', column_reversed=True), {'b': 'val2', 'a': 'val1'})

    def test_insert_get_column_count(self):
        key = 'TestColumnFamily.test_insert_get_column_count'
        columns = {'a': 'val1', 'b': 'val2', 'c': 'val3', 'd': 'val4'}
        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            ts = test_cf.insert(key, columns)
            assert_true(isinstance(ts, (int, long)))
            assert_equal(test_cf.get(key, column_count=3), {'a': 'val1', 'b': 'val2', 'c': 'val3'})

    def test_insert_get_default_column_count(self):
        keys = [str(i) for i in range(1000)]
        keys.sort()
        keys_and_values = [(key, key) for key in keys]
        key = 'TestColumnFamily.test_insert_get_default_column_count'

        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            test_cf.insert(key, dict(key_value for key_value in keys_and_values))
            assert_equal(test_cf.get(key), dict([key_value for key_value in keys_and_values][:100]))

    def test_insert_multiget(self):
        key1 = 'TestColumnFamily.test_insert_multiget1'
        columns1 = {'1': 'val1', '2': 'val2'}
        key2 = 'test_insert_multiget1'
        columns2 = {'3': 'val1', '4': 'val2'}
        missing_key = 'key3'

        for test_cf in (cf, cf_stub):
            test_cf.insert(key1, columns1)
            test_cf.insert(key2, columns2)
            rows = test_cf.multiget([key1, key2, missing_key])
            assert_equal(len(rows), 2)
            assert_equal(rows[key1], columns1)
            assert_equal(rows[key2], columns2)
            assert_true(missing_key not in rows)

    def test_insert_multiget_column_start_and_finish(self):
        key1 = 'TestColumnFamily.test_insert_multiget_column_start_and_finish1'
        columns1 = {'1': 'val1', '2': 'val2'}
        key2 = 'TestColumnFamily.test_insert_multiget_column_start_and_finish2'
        columns2 = {'3': 'val1', '4': 'val2'}
        missing_key = 'key3'

        for test_cf in (cf, cf_stub):
            test_cf.insert(key1, columns1)
            test_cf.insert(key2, columns2)
            rows = test_cf.multiget([key1, key2, missing_key], column_start='2', column_finish='3')
            assert_equal(len(rows), 2)
            assert_equal(rows[key1], {'2': 'val2'})
            assert_equal(rows[key2], {'3': 'val1'})
            assert_true(missing_key not in rows)

    def test_insert_multiget_column_finish_and_reversed(self):
        key1 = 'TestColumnFamily.test_insert_multiget_column_finish_and_reversed1'
        columns1 = {'1': 'val1', '3': 'val2'}
        key2 = 'TestColumnFamily.test_insert_multiget_column_finish_and_reversed2'
        columns2 = {'5': 'val1', '7': 'val2'}
        missing_key = 'key3'

        for test_cf in (cf, cf_stub):
            test_cf.insert(key1, columns1)
            test_cf.insert(key2, columns2)
            rows = test_cf.multiget([key1, key2, missing_key], column_finish='3', column_reversed=True)
            assert_equal(len(rows), 2)
            assert_equal(rows[key1], {'3': 'val2'})
            assert_equal(rows[key2], {'5': 'val1', '7': 'val2'})
            assert_true(missing_key not in rows)

    def test_insert_multiget_column_start_column_count(self):
        key1 = 'TestColumnFamily.test_insert_multiget_column_start_column_count'
        columns1 = {'1': 'val1', '2': 'val2'}
        key2 = 'test_insert_multiget1'
        columns2 = {'3': 'val1', '4': 'val2'}
        missing_key = 'key3'

        for test_cf in (cf, cf_stub):
            test_cf.insert(key1, columns1)
            test_cf.insert(key2, columns2)
            rows = test_cf.multiget([key1, key2, missing_key], column_count=1, column_start='2')
            assert_equal(len(rows), 2)
            assert_equal(rows[key1], {'2': 'val2'})
            assert_equal(rows[key2], {'3': 'val1'})
            assert_true(missing_key not in rows)

    def test_insert_multiget_default_column_count(self):
        keys = [str(i) for i in range(1000)]
        keys.sort()
        keys_and_values = [(key, key) for key in keys]
        key = 'TestColumnFamily.test_insert_multiget_default_column_count'

        for test_cf in (cf, cf_stub):
            test_cf.insert(key, dict(key_value for key_value in keys_and_values))
            rows = test_cf.multiget([key])
            assert_equal(len(rows), 1)
            assert_equal(rows[key], dict([key_value for key_value in keys_and_values][:100]))

    def insert_insert_get_indexed_slices(self):
        columns = {'birthdate': 1L}

        keys = set()
        for i in range(1, 4):
            indexed_cf.insert('key%d' % i, columns)
            indexed_cf_stub.insert('key%d' % i, columns)
            keys.add('key%d' % i)

        expr = index.create_index_expression(column_name='birthdate', value=1L)
        clause = index.create_index_clause([expr])

        for test_indexed_cf in (indexed_cf, indexed_cf_stub):
            count = 0
            for key, cols in test_indexed_cf.get_indexed_slices(clause):
                assert_equal(cols, columns)
                assert key in keys
                count += 1
            assert_equal(count, 3)

    def test_remove(self):
        key = 'TestColumnFamily.test_remove'
        for test_cf in (cf, cf_stub):
            columns = {'1': 'val1', '2': 'val2'}
            test_cf.insert(key, columns)

            # An empty list for columns shouldn't delete anything
            test_cf.remove(key, columns=[])
            assert_equal(test_cf.get(key), columns)

            test_cf.remove(key, columns=['2'])
            del columns['2']
            assert_equal(test_cf.get(key), {'1': 'val1'})

            test_cf.remove(key)
            assert_raises(NotFoundException, test_cf.get, key)

    def test_insert_get_tuuids(self):
        key = 'TestColumnFamily.test_insert_get'
        columns = ((convert_time_to_uuid(time.time() - 1000, randomize=True), 'val1'),
                   (convert_time_to_uuid(time.time(), randomize=True), 'val2'))
        for test_cf in (cf, cf_stub):
            assert_raises(NotFoundException, test_cf.get, key)
            ts = test_cf.insert(key, dict(columns))
            assert_true(isinstance(ts, (int, long)))
            assert_equal(test_cf.get(key).keys(), [x[0] for x in columns])



from functools import wraps

from nose.plugins.skip import SkipTest

def requireOPP(f):
    """ Decorator to require an order-preserving partitioner """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        partitioner = self.sys_man.describe_partitioner()
        if partitioner in ('RandomPartitioner', 'Murmur3Partitioner'):
            raise SkipTest('Must use order preserving partitioner for this test')
        return f(self, *args, **kwargs)

    return wrapper



import unittest

from nose import SkipTest
from nose.tools import assert_equal, assert_raises

from pycassa.pool import ConnectionPool
from pycassa.columnfamily import ColumnFamily
from pycassa.system_manager import (SIMPLE_STRATEGY, LONG_TYPE, SystemManager,
        UTF8_TYPE, TIME_UUID_TYPE, ASCII_TYPE, INT_TYPE)

from pycassa.cassandra.ttypes import InvalidRequestException
from pycassa.types import LongType

TEST_KS = 'PycassaTestKeyspace'
sys = None

def setup_module():
    global sys
    sys = SystemManager()

def teardown_module():
    sys.close()

class SystemManagerTest(unittest.TestCase):

    def test_system_calls(self):
        # keyspace modifications
        try:
            sys.drop_keyspace('TestKeyspace')
        except InvalidRequestException:
            pass
        sys.create_keyspace('TestKeyspace', SIMPLE_STRATEGY, {'replication_factor': '3'})
        sys.alter_keyspace('TestKeyspace', strategy_options={'replication_factor': '1'})

        sys.create_column_family('TestKeyspace', 'TestCF')
        sys.alter_column_family('TestKeyspace', 'TestCF', comment='testing')
        sys.create_index('TestKeyspace', 'TestCF', 'column', LONG_TYPE)
        sys.drop_column_family('TestKeyspace', 'TestCF')

        sys.describe_ring('TestKeyspace')
        sys.describe_cluster_name()
        sys.describe_version()
        sys.describe_schema_versions()
        sys.list_keyspaces()

        sys.drop_keyspace('TestKeyspace')

    def test_bad_comparator(self):
        sys.create_keyspace('TestKeyspace', SIMPLE_STRATEGY, {'replication_factor': '3'})
        for comparator in [LongType, 123]:
            assert_raises(TypeError, sys.create_column_family,
                    'TestKeyspace', 'TestBadCF', comparator_type=comparator)
        sys.drop_keyspace('TestKeyspace')

    def test_alter_column_non_bytes_type(self):
        sys.create_column_family(TEST_KS, 'LongCF', comparator_type=LONG_TYPE)
        sys.create_index(TEST_KS, 'LongCF', 3, LONG_TYPE)
        pool = ConnectionPool(TEST_KS)
        cf = ColumnFamily(pool, 'LongCF')
        cf.insert('key', {3: 3})
        assert_equal(cf.get('key')[3], 3)

        sys.alter_column(TEST_KS, 'LongCF', 2, LONG_TYPE)
        cf = ColumnFamily(pool, 'LongCF')
        cf.insert('key', {2: 2})
        assert_equal(cf.get('key')[2], 2)

    def test_alter_column_family_default_validation_class(self):
        sys.create_column_family(TEST_KS, 'AlteredCF', default_validation_class=LONG_TYPE)
        pool = ConnectionPool(TEST_KS)
        cf = ColumnFamily(pool, 'AlteredCF')
        assert_equal(cf.default_validation_class, "LongType")

        sys.alter_column_family(TEST_KS, 'AlteredCF', default_validation_class=UTF8_TYPE)
        cf = ColumnFamily(pool, 'AlteredCF')
        assert_equal(cf.default_validation_class, "UTF8Type")

    def test_alter_column_super_cf(self):
        sys.create_column_family(TEST_KS, 'SuperCF', super=True,
                comparator_type=TIME_UUID_TYPE, subcomparator_type=UTF8_TYPE)
        sys.alter_column(TEST_KS, 'SuperCF', 'foobar_col', UTF8_TYPE)

    def test_column_validators(self):
        validators = {'name': UTF8_TYPE, 'age': LONG_TYPE}
        sys.create_column_family(TEST_KS, 'ValidatedCF',
                column_validation_classes=validators)
        pool = ConnectionPool(TEST_KS)
        cf = ColumnFamily(pool, 'ValidatedCF')
        cf.insert('key', {'name': 'John', 'age': 40})
        self.assertEquals(cf.get('key'), {'name': 'John', 'age': 40})

        validators = {'name': ASCII_TYPE, 'age': INT_TYPE}
        sys.alter_column_family(TEST_KS, 'ValidatedCF',
                column_validation_classes=validators)
        cf.load_schema()
        self.assertEquals(cf.get('key'), {'name': 'John', 'age': 40})

    def test_caching_pre_11(self):
        version = tuple(
            [int(v) for v in sys._conn.describe_version().split('.')])
        if version >= (19, 30, 0):
            raise SkipTest('CF specific caching no longer supported.')
        sys.create_column_family(TEST_KS, 'CachedCF10',
            row_cache_size=100, key_cache_size=100,
            row_cache_save_period_in_seconds=3,
            key_cache_save_period_in_seconds=3)
        pool = ConnectionPool(TEST_KS)
        cf = ColumnFamily(pool, 'CachedCF10')
        assert_equal(cf._cfdef.row_cache_size, 100)
        assert_equal(cf._cfdef.key_cache_size, 100)
        assert_equal(cf._cfdef.row_cache_save_period_in_seconds, 3)
        assert_equal(cf._cfdef.key_cache_save_period_in_seconds, 3)
        sys.alter_column_family(TEST_KS, 'CachedCF10',
            row_cache_size=200, key_cache_size=200,
            row_cache_save_period_in_seconds=4,
            key_cache_save_period_in_seconds=4)
        cf1 = ColumnFamily(pool, 'CachedCF10')
        assert_equal(cf1._cfdef.row_cache_size, 200)
        assert_equal(cf1._cfdef.key_cache_size, 200)
        assert_equal(cf1._cfdef.row_cache_save_period_in_seconds, 4)
        assert_equal(cf1._cfdef.key_cache_save_period_in_seconds, 4)

    def test_caching_post_11(self):
        version = tuple(
            [int(v) for v in sys._conn.describe_version().split('.')])
        if version < (19, 30, 0):
            raise SkipTest('CF caching policy not yet supported.')
        sys.create_column_family(TEST_KS, 'CachedCF11')
        pool = ConnectionPool(TEST_KS)
        cf = ColumnFamily(pool, 'CachedCF11')
        assert_equal(cf._cfdef.caching, 'KEYS_ONLY')
        sys.alter_column_family(TEST_KS, 'CachedCF11', caching='all')
        cf = ColumnFamily(pool, 'CachedCF11')
        assert_equal(cf._cfdef.caching, 'ALL')
        sys.alter_column_family(TEST_KS, 'CachedCF11', caching='rows_only')
        cf = ColumnFamily(pool, 'CachedCF11')
        assert_equal(cf._cfdef.caching, 'ROWS_ONLY')
        sys.alter_column_family(TEST_KS, 'CachedCF11', caching='none')
        cf = ColumnFamily(pool, 'CachedCF11')
        assert_equal(cf._cfdef.caching, 'NONE')



from pycassa import NotFoundException
from pycassa.pool import ConnectionPool
from pycassa.columnfamily import ColumnFamily
from pycassa.util import OrderedDict, convert_uuid_to_time
from pycassa.system_manager import SystemManager
from pycassa.types import (LongType, IntegerType, TimeUUIDType, LexicalUUIDType,
                           AsciiType, UTF8Type, BytesType, CompositeType,
                           OldPycassaDateType, IntermediateDateType, DateType,
                           BooleanType, CassandraType, DecimalType,
                           FloatType, Int32Type, UUIDType, DoubleType, DynamicCompositeType)
from pycassa.index import create_index_expression, create_index_clause
import pycassa.marshal as marshal

from nose import SkipTest
from nose.tools import (assert_raises, assert_equal, assert_almost_equal,
                        assert_true)

from datetime import date, datetime
from uuid import uuid1
from decimal import Decimal
import uuid
import unittest
import time
from collections import namedtuple

TIME1 = uuid.UUID(hex='ddc6118e-a003-11df-8abf-00234d21610a')
TIME2 = uuid.UUID(hex='40ad6d4c-a004-11df-8abf-00234d21610a')
TIME3 = uuid.UUID(hex='dc3d5234-a00b-11df-8abf-00234d21610a')

VALS = ['val1', 'val2', 'val3']
KEYS = ['key1', 'key2', 'key3']

pool = None
TEST_KS = 'PycassaTestKeyspace'

def setup_module():
    global pool
    credentials = {'username': 'jsmith', 'password': 'havebadpass'}
    pool = ConnectionPool(TEST_KS, pool_size=10, credentials=credentials, timeout=1.0)

def teardown_module():
    pool.dispose()

class TestCFs(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'StdLong', comparator_type=LongType())
        sys.create_column_family(TEST_KS, 'StdInteger', comparator_type=IntegerType())
        sys.create_column_family(TEST_KS, 'StdBigInteger', comparator_type=IntegerType())
        sys.create_column_family(TEST_KS, 'StdDecimal', comparator_type=DecimalType())
        sys.create_column_family(TEST_KS, 'StdTimeUUID', comparator_type=TimeUUIDType())
        sys.create_column_family(TEST_KS, 'StdLexicalUUID', comparator_type=LexicalUUIDType())
        sys.create_column_family(TEST_KS, 'StdAscii', comparator_type=AsciiType())
        sys.create_column_family(TEST_KS, 'StdUTF8', comparator_type=UTF8Type())
        sys.create_column_family(TEST_KS, 'StdBytes', comparator_type=BytesType())
        sys.create_column_family(TEST_KS, 'StdComposite',
                                 comparator_type=CompositeType(LongType(), BytesType()))
        sys.create_column_family(TEST_KS, 'StdDynamicComposite',
                                 comparator_type=DynamicCompositeType({'a': AsciiType(),
                                 'b': BytesType(), 'c': DecimalType(), 'd': DateType(),
                                 'f': FloatType(), 'i': IntegerType(), 'l': LongType(),
                                 'n': Int32Type(), 's': UTF8Type(), 't': TimeUUIDType(),
                                 'u': UUIDType(), 'w': DoubleType(), 'x': LexicalUUIDType(),
                                 'y': BooleanType()}))
        sys.close()

        cls.cf_long = ColumnFamily(pool, 'StdLong')
        cls.cf_int = ColumnFamily(pool, 'StdInteger')
        cls.cf_big_int = ColumnFamily(pool, 'StdBigInteger')
        cls.cf_decimal = ColumnFamily(pool, 'StdDecimal')
        cls.cf_time = ColumnFamily(pool, 'StdTimeUUID')
        cls.cf_lex = ColumnFamily(pool, 'StdLexicalUUID')
        cls.cf_ascii = ColumnFamily(pool, 'StdAscii')
        cls.cf_utf8 = ColumnFamily(pool, 'StdUTF8')
        cls.cf_bytes = ColumnFamily(pool, 'StdBytes')
        cls.cf_composite = ColumnFamily(pool, 'StdComposite')
        cls.cf_dynamic_composite = ColumnFamily(pool, 'StdDynamicComposite')

        cls.cfs = [cls.cf_long, cls.cf_int, cls.cf_time, cls.cf_lex,
                   cls.cf_ascii, cls.cf_utf8, cls.cf_bytes, cls.cf_composite, 
                   cls.cf_dynamic_composite]

    def tearDown(self):
        for cf in TestCFs.cfs:
            for key, cols in cf.get_range():
                cf.remove(key)

    def make_group(self, cf, cols):
        diction = OrderedDict([(cols[0], VALS[0]),
                               (cols[1], VALS[1]),
                               (cols[2], VALS[2])])
        return {'cf': cf, 'cols': cols, 'dict': diction}

    def test_standard_column_family(self):

        # For each data type, create a group that includes its column family,
        # a set of column names, and a dictionary that maps from the column
        # names to values.
        type_groups = []

        long_cols = [1111111111111111L,
                     2222222222222222L,
                     3333333333333333L]
        type_groups.append(self.make_group(TestCFs.cf_long, long_cols))

        int_cols = [1, 2, 3]
        type_groups.append(self.make_group(TestCFs.cf_int, int_cols))

        big_int_cols = [1 + int(time.time() * 10 ** 6),
                        2 + int(time.time() * 10 ** 6),
                        3 + int(time.time() * 10 ** 6)]
        type_groups.append(self.make_group(TestCFs.cf_big_int, big_int_cols))

        decimal_cols = [Decimal('1.123456789123456789'),
                        Decimal('2.123456789123456789'),
                        Decimal('3.123456789123456789')]
        type_groups.append(self.make_group(TestCFs.cf_decimal, decimal_cols))

        time_cols = [TIME1, TIME2, TIME3]
        type_groups.append(self.make_group(TestCFs.cf_time, time_cols))

        lex_cols = [uuid.UUID(bytes='aaa aaa aaa aaaa'),
                    uuid.UUID(bytes='bbb bbb bbb bbbb'),
                    uuid.UUID(bytes='ccc ccc ccc cccc')]
        type_groups.append(self.make_group(TestCFs.cf_lex, lex_cols))

        ascii_cols = ['aaaa', 'bbbb', 'cccc']
        type_groups.append(self.make_group(TestCFs.cf_ascii, ascii_cols))

        utf8_cols = [u'a\u0020', u'b\u0020', u'c\u0020']
        type_groups.append(self.make_group(TestCFs.cf_utf8, utf8_cols))

        bytes_cols = ['aaaa', 'bbbb', 'cccc']
        type_groups.append(self.make_group(TestCFs.cf_bytes, bytes_cols))

        composite_cols = [(1, 'foo'), (2, 'bar'), (3, 'baz')]
        type_groups.append(self.make_group(TestCFs.cf_composite, composite_cols))

        dynamic_composite_cols = [(('LongType', 1), ('BytesType', 'foo')), 
                                  (('LongType', 2), ('BytesType', 'bar')), 
                                  (('LongType', 3), ('BytesType', 'baz'))]
        type_groups.append(self.make_group(TestCFs.cf_dynamic_composite, dynamic_composite_cols))

        dynamic_composite_alias_cols = [(('l', 1), ('b', 'foo')), 
                                        (('l', 2), ('b', 'bar')), 
                                        (('l', 3), ('b', 'baz'))]
        type_groups.append(self.make_group(TestCFs.cf_dynamic_composite, dynamic_composite_alias_cols))

        # Begin the actual inserting and getting	
        for group in type_groups:
            cf = group.get('cf')
            gdict = group.get('dict')
            gcols = group.get('cols')

            cf.insert(KEYS[0], gdict)
            assert_equal(cf.get(KEYS[0]), gdict)

            # Check each column individually
            for i in range(3):
                assert_equal(cf.get(KEYS[0], columns=[gcols[i]]),
                             {gcols[i]: VALS[i]})

            # Check that if we list all columns, we get the full dict
            assert_equal(cf.get(KEYS[0], columns=gcols[:]), gdict)
            # The same thing with a start and end instead
            assert_equal(cf.get(KEYS[0], column_start=gcols[0], column_finish=gcols[2]),
                         gdict)
            # A start and end that are the same
            assert_equal(cf.get(KEYS[0], column_start=gcols[0], column_finish=gcols[0]),
                         {gcols[0]: VALS[0]})

            assert_equal(cf.get_count(KEYS[0]), 3)

            # Test xget paging
            assert_equal(list(cf.xget(KEYS[0], buffer_size=2)), gdict.items())
            assert_equal(list(cf.xget(KEYS[0], column_reversed=True, buffer_size=2)),
                         list(reversed(gdict.items())))
            assert_equal(list(cf.xget(KEYS[0], column_start=gcols[0], buffer_size=2)),
                         gdict.items())
            assert_equal(list(cf.xget(KEYS[0], column_finish=gcols[2], buffer_size=2)),
                         gdict.items())
            assert_equal(list(cf.xget(KEYS[0], column_start=gcols[2], column_finish=gcols[0],
                                      column_reversed=True, buffer_size=2)),
                         list(reversed(gdict.items())))
            assert_equal(list(cf.xget(KEYS[0], column_start=gcols[1], column_finish=gcols[1],
                                      column_reversed=True, buffer_size=2)),
                         [(gcols[1], VALS[1])])

            # Test removing rows
            cf.remove(KEYS[0], columns=gcols[:1])
            assert_equal(cf.get_count(KEYS[0]), 2)

            cf.remove(KEYS[0], columns=gcols[1:])
            assert_equal(cf.get_count(KEYS[0]), 0)

            # Insert more than one row now
            cf.insert(KEYS[0], gdict)
            cf.insert(KEYS[1], gdict)
            cf.insert(KEYS[2], gdict)

            ### multiget() tests ###

            res = cf.multiget(KEYS[:])
            for i in range(3):
                assert_equal(res.get(KEYS[i]), gdict)

            res = cf.multiget(KEYS[2:])
            assert_equal(res.get(KEYS[2]), gdict)

            # Check each column individually
            for i in range(3):
                res = cf.multiget(KEYS[:], columns=[gcols[i]])
                for j in range(3):
                    assert_equal(res.get(KEYS[j]), {gcols[i]: VALS[i]})

            # Check that if we list all columns, we get the full dict
            res = cf.multiget(KEYS[:], columns=gcols[:])
            for j in range(3):
                assert_equal(res.get(KEYS[j]), gdict)

            # The same thing with a start and end instead
            res = cf.multiget(KEYS[:], column_start=gcols[0], column_finish=gcols[2])
            for j in range(3):
                assert_equal(res.get(KEYS[j]), gdict)

            # A start and end that are the same
            res = cf.multiget(KEYS[:], column_start=gcols[0], column_finish=gcols[0])
            for j in range(3):
                assert_equal(res.get(KEYS[j]), {gcols[0]: VALS[0]})

            ### get_range() tests ###

            res = cf.get_range(start=KEYS[0])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], column_start=gcols[0], column_finish=gcols[2])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], columns=gcols[:])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)


class TestSuperCFs(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'SuperLong', super=True, comparator_type=LongType())
        sys.create_column_family(TEST_KS, 'SuperInt', super=True, comparator_type=IntegerType())
        sys.create_column_family(TEST_KS, 'SuperBigInt', super=True, comparator_type=IntegerType())
        sys.create_column_family(TEST_KS, 'SuperTime', super=True, comparator_type=TimeUUIDType())
        sys.create_column_family(TEST_KS, 'SuperLex', super=True, comparator_type=LexicalUUIDType())
        sys.create_column_family(TEST_KS, 'SuperAscii', super=True, comparator_type=AsciiType())
        sys.create_column_family(TEST_KS, 'SuperUTF8', super=True, comparator_type=UTF8Type())
        sys.create_column_family(TEST_KS, 'SuperBytes', super=True, comparator_type=BytesType())
        sys.close()

        cls.cf_suplong = ColumnFamily(pool, 'SuperLong')
        cls.cf_supint = ColumnFamily(pool, 'SuperInt')
        cls.cf_supbigint = ColumnFamily(pool, 'SuperBigInt')
        cls.cf_suptime = ColumnFamily(pool, 'SuperTime')
        cls.cf_suplex = ColumnFamily(pool, 'SuperLex')
        cls.cf_supascii = ColumnFamily(pool, 'SuperAscii')
        cls.cf_suputf8 = ColumnFamily(pool, 'SuperUTF8')
        cls.cf_supbytes = ColumnFamily(pool, 'SuperBytes')

        cls.cfs = [cls.cf_suplong, cls.cf_supint, cls.cf_suptime,
                   cls.cf_suplex, cls.cf_supascii, cls.cf_suputf8,
                   cls.cf_supbytes]

    def tearDown(self):
        for cf in TestSuperCFs.cfs:
            for key, cols in cf.get_range():
                cf.remove(key)

    def make_super_group(self, cf, cols):
        diction = OrderedDict([(cols[0], {'bytes': VALS[0]}),
                               (cols[1], {'bytes': VALS[1]}),
                               (cols[2], {'bytes': VALS[2]})])
        return {'cf': cf, 'cols': cols, 'dict': diction}

    def test_super_column_families(self):

        # For each data type, create a group that includes its column family,
        # a set of column names, and a dictionary that maps from the column
        # names to values.
        type_groups = []

        long_cols = [1111111111111111L,
                     2222222222222222L,
                     3333333333333333L]
        type_groups.append(self.make_super_group(TestSuperCFs.cf_suplong, long_cols))

        int_cols = [1, 2, 3]
        type_groups.append(self.make_super_group(TestSuperCFs.cf_supint, int_cols))

        big_int_cols = [1 + int(time.time() * 10 ** 6),
                        2 + int(time.time() * 10 ** 6),
                        3 + int(time.time() * 10 ** 6)]
        type_groups.append(self.make_super_group(TestSuperCFs.cf_supbigint, big_int_cols))

        time_cols = [TIME1, TIME2, TIME3]
        type_groups.append(self.make_super_group(TestSuperCFs.cf_suptime, time_cols))

        lex_cols = [uuid.UUID(bytes='aaa aaa aaa aaaa'),
                    uuid.UUID(bytes='bbb bbb bbb bbbb'),
                    uuid.UUID(bytes='ccc ccc ccc cccc')]
        type_groups.append(self.make_super_group(TestSuperCFs.cf_suplex, lex_cols))

        ascii_cols = ['aaaa', 'bbbb', 'cccc']
        type_groups.append(self.make_super_group(TestSuperCFs.cf_supascii, ascii_cols))

        utf8_cols = [u'a\u0020', u'b\u0020', u'c\u0020']
        type_groups.append(self.make_super_group(TestSuperCFs.cf_suputf8, utf8_cols))

        bytes_cols = ['aaaa', 'bbbb', 'cccc']
        type_groups.append(self.make_super_group(TestSuperCFs.cf_supbytes, bytes_cols))

        # Begin the actual inserting and getting
        for group in type_groups:
            cf = group.get('cf')
            gdict = group.get('dict')
            gcols = group.get('cols')

            cf.insert(KEYS[0], gdict)
            assert_equal(cf.get(KEYS[0]), gdict)

            # Check each supercolumn individually
            for i in range(3):
                res = cf.get(KEYS[0], columns=[gcols[i]])
                assert_equal(res, {gcols[i]: {'bytes': VALS[i]}})

            # Check that if we list all columns, we get the full dict
            assert_equal(cf.get(KEYS[0], columns=gcols[:]), gdict)
            # The same thing with a start and end instead
            assert_equal(cf.get(KEYS[0], column_start=gcols[0], column_finish=gcols[2]), gdict)
            # A start and end that are the same
            assert_equal(cf.get(KEYS[0], column_start=gcols[0], column_finish=gcols[0]),
                         {gcols[0]: {'bytes': VALS[0]}})

            # test xget paging
            assert_equal(list(cf.xget(KEYS[0], buffer_size=2)), gdict.items())

            assert_equal(cf.get_count(KEYS[0]), 3)

            # Test removing rows
            cf.remove(KEYS[0], columns=gcols[:1])
            assert_equal(cf.get_count(KEYS[0]), 2)

            cf.remove(KEYS[0], columns=gcols[1:])
            assert_equal(cf.get_count(KEYS[0]), 0)

            # Insert more than one row now
            cf.insert(KEYS[0], gdict)
            cf.insert(KEYS[1], gdict)
            cf.insert(KEYS[2], gdict)

            ### multiget() tests ###

            res = cf.multiget(KEYS[:])
            for i in range(3):
                assert_equal(res.get(KEYS[i]), gdict)

            res = cf.multiget(KEYS[2:])
            assert_equal(res.get(KEYS[2]), gdict)

            # Check each column individually
            for i in range(3):
                res = cf.multiget(KEYS[:], columns=[gcols[i]])
                for j in range(3):
                    assert_equal(res.get(KEYS[j]), {gcols[i]: {'bytes': VALS[i]}})

            # Check that if we list all columns, we get the full dict
            res = cf.multiget(KEYS[:], columns=gcols[:])
            for j in range(3):
                assert_equal(res.get(KEYS[j]), gdict)

            # The same thing with a start and end instead
            res = cf.multiget(KEYS[:], column_start=gcols[0], column_finish=gcols[2])
            for j in range(3):
                assert_equal(res.get(KEYS[j]), gdict)

            # A start and end that are the same
            res = cf.multiget(KEYS[:], column_start=gcols[0], column_finish=gcols[0])
            for j in range(3):
                assert_equal(res.get(KEYS[j]), {gcols[0]: {'bytes': VALS[0]}})

            ### get_range() tests ###

            res = cf.get_range(start=KEYS[0])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], column_start=gcols[0], column_finish=gcols[2])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], columns=gcols[:])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)


class TestSuperSubCFs(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'SuperLongSubLong', super=True,
                                 comparator_type=LongType(), subcomparator_type=LongType())
        sys.create_column_family(TEST_KS, 'SuperLongSubInt', super=True,
                                 comparator_type=LongType(), subcomparator_type=IntegerType())
        sys.create_column_family(TEST_KS, 'SuperLongSubBigInt', super=True,
                                 comparator_type=LongType(), subcomparator_type=IntegerType())
        sys.create_column_family(TEST_KS, 'SuperLongSubTime', super=True,
                                 comparator_type=LongType(), subcomparator_type=TimeUUIDType())
        sys.create_column_family(TEST_KS, 'SuperLongSubLex', super=True,
                                 comparator_type=LongType(), subcomparator_type=LexicalUUIDType())
        sys.create_column_family(TEST_KS, 'SuperLongSubAscii', super=True,
                                 comparator_type=LongType(), subcomparator_type=AsciiType())
        sys.create_column_family(TEST_KS, 'SuperLongSubUTF8', super=True,
                                 comparator_type=LongType(), subcomparator_type=UTF8Type())
        sys.create_column_family(TEST_KS, 'SuperLongSubBytes', super=True,
                                 comparator_type=LongType(), subcomparator_type=BytesType())
        sys.close()

        cls.cf_suplong_sublong = ColumnFamily(pool, 'SuperLongSubLong')
        cls.cf_suplong_subint = ColumnFamily(pool, 'SuperLongSubInt')
        cls.cf_suplong_subbigint = ColumnFamily(pool, 'SuperLongSubBigInt')
        cls.cf_suplong_subtime = ColumnFamily(pool, 'SuperLongSubTime')
        cls.cf_suplong_sublex = ColumnFamily(pool, 'SuperLongSubLex')
        cls.cf_suplong_subascii = ColumnFamily(pool, 'SuperLongSubAscii')
        cls.cf_suplong_subutf8 = ColumnFamily(pool, 'SuperLongSubUTF8')
        cls.cf_suplong_subbytes = ColumnFamily(pool, 'SuperLongSubBytes')

        cls.cfs = [cls.cf_suplong_subint, cls.cf_suplong_subint,
                   cls.cf_suplong_subtime, cls.cf_suplong_sublex,
                   cls.cf_suplong_subascii, cls.cf_suplong_subutf8,
                   cls.cf_suplong_subbytes]

    def tearDown(self):
        for cf in TestSuperSubCFs.cfs:
            for key, cols in cf.get_range():
                cf.remove(key)

    def make_sub_group(self, cf, cols):
        diction = {123L: {cols[0]: VALS[0],
                          cols[1]: VALS[1],
                          cols[2]: VALS[2]}}
        return {'cf': cf, 'cols': cols, 'dict': diction}

    def test_super_column_family_subs(self):

        # For each data type, create a group that includes its column family,
        # a set of column names, and a dictionary that maps from the column
        # names to values.
        type_groups = []

        long_cols = [1111111111111111L,
                     2222222222222222L,
                     3333333333333333L]
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_sublong, long_cols))

        int_cols = [1, 2, 3]
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_subint, int_cols))

        big_int_cols = [1 + int(time.time() * 10 ** 6),
                        2 + int(time.time() * 10 ** 6),
                        3 + int(time.time() * 10 ** 6)]
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_subbigint, big_int_cols))

        time_cols = [TIME1, TIME2, TIME3]
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_subtime, time_cols))

        lex_cols = [uuid.UUID(bytes='aaa aaa aaa aaaa'),
                    uuid.UUID(bytes='bbb bbb bbb bbbb'),
                    uuid.UUID(bytes='ccc ccc ccc cccc')]
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_sublex, lex_cols))

        ascii_cols = ['aaaa', 'bbbb', 'cccc']
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_subascii, ascii_cols))

        utf8_cols = [u'a\u0020', u'b\u0020', u'c\u0020']
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_subutf8, utf8_cols))

        bytes_cols = ['aaaa', 'bbbb', 'cccc']
        type_groups.append(self.make_sub_group(TestSuperSubCFs.cf_suplong_subbytes, bytes_cols))

        # Begin the actual inserting and getting
        for group in type_groups:
            cf = group.get('cf')
            gdict = group.get('dict')

            cf.insert(KEYS[0], gdict)

            assert_equal(cf.get(KEYS[0]), gdict)
            assert_equal(cf.get(KEYS[0], columns=[123L]), gdict)

            # A start and end that are the same
            assert_equal(cf.get(KEYS[0], column_start=123L, column_finish=123L), gdict)

            res = cf.get(KEYS[0], super_column=123L, column_start=group.get('cols')[0])
            assert_equal(res, gdict.get(123L))

            res = cf.get(KEYS[0], super_column=123L, column_finish=group.get('cols')[-1])
            assert_equal(res, gdict.get(123L))

            assert_equal(cf.get_count(KEYS[0]), 1)

            # Test removing rows
            cf.remove(KEYS[0], super_column=123L)
            assert_equal(cf.get_count(KEYS[0]), 0)

            # Insert more than one row now
            cf.insert(KEYS[0], gdict)
            cf.insert(KEYS[1], gdict)
            cf.insert(KEYS[2], gdict)

            ### multiget() tests ###

            res = cf.multiget(KEYS[:])
            for i in range(3):
                assert_equal(res.get(KEYS[i]), gdict)

            res = cf.multiget(KEYS[2:])
            assert_equal(res.get(KEYS[2]), gdict)

            res = cf.multiget(KEYS[:], columns=[123L])
            for i in range(3):
                assert_equal(res.get(KEYS[i]), gdict)

            res = cf.multiget(KEYS[:], super_column=123L)
            for i in range(3):
                assert_equal(res.get(KEYS[i]), gdict.get(123L))

            res = cf.multiget(KEYS[:], super_column=123L, column_start=group.get('cols')[0])
            for i in range(3):
                assert_equal(res.get(KEYS[i]), gdict.get(123L))

            res = cf.multiget(KEYS[:], column_start=123L, column_finish=123L)
            for j in range(3):
                assert_equal(res.get(KEYS[j]), gdict)

            ### get_range() tests ###

            res = cf.get_range(start=KEYS[0])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], column_start=123L, column_finish=123L)
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], columns=[123L])
            for sub_res in res:
                assert_equal(sub_res[1], gdict)

            res = cf.get_range(start=KEYS[0], super_column=123L)
            for sub_res in res:
                assert_equal(sub_res[1], gdict.get(123L))

            res = cf.get_range(start=KEYS[0], super_column=123L, column_start=group.get('cols')[0])
            for sub_res in res:
                assert_equal(sub_res[1], gdict.get(123L))

            res = cf.get_range(start=KEYS[0], super_column=123L, column_finish=group.get('cols')[-1])
            for sub_res in res:
                assert_equal(sub_res[1], gdict.get(123L))

class TestValidators(unittest.TestCase):

    def test_validation_with_packed_names(self):
        """
        Make sure that validated columns are packed correctly when the
        column names themselves must be packed
        """
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'Validators2',
                comparator_type=LongType(), default_validation_class=LongType())
        sys.alter_column(TEST_KS, 'Validators2', 1, TimeUUIDType())
        sys.close()

        my_uuid = uuid.uuid1()
        cf = ColumnFamily(pool, 'Validators2')

        cf.insert('key', {0: 0})
        assert_equal(cf.get('key'), {0: 0})

        cf.insert('key', {1: my_uuid})
        assert_equal(cf.get('key'), {0: 0, 1: my_uuid})

        cf.insert('key', {0: 0, 1: my_uuid})
        assert_equal(cf.get('key'), {0: 0, 1: my_uuid})

    def test_validated_columns(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'Validators',)
        sys.alter_column(TEST_KS, 'Validators', 'long', LongType())
        sys.alter_column(TEST_KS, 'Validators', 'int', IntegerType())
        sys.alter_column(TEST_KS, 'Validators', 'time', TimeUUIDType())
        sys.alter_column(TEST_KS, 'Validators', 'lex', LexicalUUIDType())
        sys.alter_column(TEST_KS, 'Validators', 'ascii', AsciiType())
        sys.alter_column(TEST_KS, 'Validators', 'utf8', UTF8Type())
        sys.alter_column(TEST_KS, 'Validators', 'bytes', BytesType())
        sys.close()

        cf = ColumnFamily(pool, 'Validators')
        key = 'key1'

        col = {'long': 1L}
        cf.insert(key, col)
        assert_equal(cf.get(key)['long'], 1L)

        col = {'int': 1}
        cf.insert(key, col)
        assert_equal(cf.get(key)['int'], 1)

        col = {'time': TIME1}
        cf.insert(key, col)
        assert_equal(cf.get(key)['time'], TIME1)

        col = {'lex': uuid.UUID(bytes='aaa aaa aaa aaaa')}
        cf.insert(key, col)
        assert_equal(cf.get(key)['lex'], uuid.UUID(bytes='aaa aaa aaa aaaa'))

        col = {'ascii': 'aaa'}
        cf.insert(key, col)
        assert_equal(cf.get(key)['ascii'], 'aaa')

        col = {'utf8': u'a\u0020'}
        cf.insert(key, col)
        assert_equal(cf.get(key)['utf8'], u'a\u0020')

        col = {'bytes': 'aaa'}
        cf.insert(key, col)
        assert_equal(cf.get(key)['bytes'], 'aaa')

        cf.remove(key)


class TestDefaultValidators(unittest.TestCase):

    def test_default_validated_columns(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'DefaultValidator', default_validation_class=LongType())
        sys.alter_column(TEST_KS, 'DefaultValidator', 'subcol', TimeUUIDType())
        sys.close()

        cf = ColumnFamily(pool, 'DefaultValidator')
        key = 'key1'

        col_cf = {'aaaaaa': 1L}
        col_cm = {'subcol': TIME1}
        col_ncf = {'aaaaaa': TIME1}

        # Both of these inserts work, as cf allows
        #  longs and cm for 'subcol' allows TIMEUUIDs.
        cf.insert(key, col_cf)
        cf.insert(key, col_cm)
        assert_equal(cf.get(key), {'aaaaaa': 1L, 'subcol': TIME1})

        # Insert multiple columns at once
        col_cf.update(col_cm)
        cf.insert(key, col_cf)
        assert_equal(cf.get(key), {'aaaaaa': 1L, 'subcol': TIME1})

        assert_raises(TypeError, cf.insert, key, col_ncf)

        cf.remove(key)

class TestKeyValidators(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()

        sys.create_column_family(TEST_KS, 'KeyLong', key_validation_class=LongType())
        sys.create_column_family(TEST_KS, 'KeyInteger', key_validation_class=IntegerType())
        sys.create_column_family(TEST_KS, 'KeyTimeUUID', key_validation_class=TimeUUIDType())
        sys.create_column_family(TEST_KS, 'KeyLexicalUUID', key_validation_class=LexicalUUIDType())
        sys.create_column_family(TEST_KS, 'KeyAscii', key_validation_class=AsciiType())
        sys.create_column_family(TEST_KS, 'KeyUTF8', key_validation_class=UTF8Type())
        sys.create_column_family(TEST_KS, 'KeyBytes', key_validation_class=BytesType())
        sys.close()

        cls.cf_long = ColumnFamily(pool, 'KeyLong')
        cls.cf_int = ColumnFamily(pool, 'KeyInteger')
        cls.cf_time = ColumnFamily(pool, 'KeyTimeUUID')
        cls.cf_lex = ColumnFamily(pool, 'KeyLexicalUUID')
        cls.cf_ascii = ColumnFamily(pool, 'KeyAscii')
        cls.cf_utf8 = ColumnFamily(pool, 'KeyUTF8')
        cls.cf_bytes = ColumnFamily(pool, 'KeyBytes')

        cls.cfs = [cls.cf_long, cls.cf_int, cls.cf_time, cls.cf_lex,
                    cls.cf_ascii, cls.cf_utf8, cls.cf_bytes]

    def tearDown(self):
        for cf in TestKeyValidators.cfs:
            for key, cols in cf.get_range():
                cf.remove(key)

    def setUp(self):
        self.type_groups = []

        long_keys = [1111111111111111L,
                     2222222222222222L,
                     3333333333333333L]
        self.type_groups.append((TestKeyValidators.cf_long, long_keys))

        int_keys = [1, 2, 3]
        self.type_groups.append((TestKeyValidators.cf_int, int_keys))

        time_keys = [TIME1, TIME2, TIME3]
        self.type_groups.append((TestKeyValidators.cf_time, time_keys))

        lex_keys = [uuid.UUID(bytes='aaa aaa aaa aaaa'),
                    uuid.UUID(bytes='bbb bbb bbb bbbb'),
                    uuid.UUID(bytes='ccc ccc ccc cccc')]
        self.type_groups.append((TestKeyValidators.cf_lex, lex_keys))

        ascii_keys = ['aaaa', 'bbbb', 'cccc']
        self.type_groups.append((TestKeyValidators.cf_ascii, ascii_keys))

        utf8_keys = [u'a\u0020', u'b\u0020', u'c\u0020']
        self.type_groups.append((TestKeyValidators.cf_utf8, utf8_keys))

        bytes_keys = ['aaaa', 'bbbb', 'cccc']
        self.type_groups.append((TestKeyValidators.cf_bytes, bytes_keys))

    def test_inserts(self):
        for cf, keys in self.type_groups:
            for key in keys:
                cf.insert(key, {str(key): 'val'})
                results = cf.get(key)
                assert_equal(results, {str(key): 'val'})

                col1 = str(key) + "1"
                col2 = str(key) + "2"
                cols = {col1: "val1", col2: "val2"}
                cf.insert(key, cols)
                results = cf.get(key)
                cols.update({str(key): 'val'})
                assert_equal(results, cols)

    def test_batch_insert(self):
        for cf, keys in self.type_groups:
            rows = dict([(key, {str(key): 'val'}) for key in keys])
            cf.batch_insert(rows)
            for key in keys:
                results = cf.get(key)
                assert_equal(results, {str(key): 'val'})

    def test_multiget(self):
        for cf, keys in self.type_groups:
            for key in keys:
                cf.insert(key, {str(key): 'val'})
            results = cf.multiget(keys)
            for key in keys:
                assert_true(key in results)
                assert_equal(results[key], {str(key): 'val'})

    def test_get_count(self):
        for cf, keys in self.type_groups:
            for key in keys:
                cf.insert(key, {str(key): 'val'})
                results = cf.get_count(key)
                assert_equal(results, 1)

    def test_multiget_count(self):
        for cf, keys in self.type_groups:
            for key in keys:
                cf.insert(key, {str(key): 'val'})
            results = cf.multiget_count(keys)
            for key in keys:
                assert_true(key in results, "%s should be in %r" % (key, results))
                assert_equal(results[key], 1)

    def test_get_range(self):
        for cf, keys in self.type_groups:
            for key in keys:
                cf.insert(key, {str(key): 'val'})

            rows = list(cf.get_range())
            assert_equal(len(rows), len(keys))
            for k, c in rows:
                assert_true(k in keys)
                assert_equal(c, {str(k): 'val'})

    def test_get_indexed_slices(self):
        sys = SystemManager()
        for cf, keys in self.type_groups:
            sys.create_index(TEST_KS, cf.column_family, 'birthdate', LongType())
            cf = ColumnFamily(pool, cf.column_family)
            for key in keys:
                cf.insert(key, {'birthdate': 1})
            expr = create_index_expression('birthdate', 1)
            clause = create_index_clause([expr])
            rows = list(cf.get_indexed_slices(clause))
            assert_equal(len(rows), len(keys))
            for k, c in rows:
                assert_true(k in keys)
                assert_equal(c, {'birthdate': 1})

    def test_remove(self):
        for cf, keys in self.type_groups:
            for key in keys:
                cf.insert(key, {str(key): 'val'})
                assert_equal(cf.get(key), {str(key): 'val'})
                cf.remove(key)
                assert_raises(NotFoundException, cf.get, key)

    def test_add_remove_counter(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'KeyLongCounter', key_validation_class=LongType(),
                                 default_validation_class='CounterColumnType')
        sys.close()
        cf_long = ColumnFamily(pool, 'KeyLongCounter')

        key = 1111111111111111L

        cf_long.add(key, 'col')
        assert_equal(cf_long.get(key), {'col': 1})
        cf_long.remove_counter(key, 'col')
        time.sleep(0.1)
        assert_raises(NotFoundException, cf_long.get, key)

class TestComposites(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'StaticComposite',
                                 comparator_type=CompositeType(LongType(),
                                                               IntegerType(),
                                                               TimeUUIDType(reversed=True),
                                                               LexicalUUIDType(reversed=False),
                                                               AsciiType(),
                                                               UTF8Type(),
                                                               BytesType()))

    @classmethod
    def teardown_class(cls):
        sys = SystemManager()
        sys.drop_column_family(TEST_KS, 'StaticComposite')

    def test_static_composite_basic(self):
        cf = ColumnFamily(pool, 'StaticComposite')
        colname = (127312831239123123, 1, uuid.uuid1(), uuid.uuid4(), 'foo', u'ba\u0254r', 'baz')
        cf.insert('key', {colname: 'val'})
        assert_equal(cf.get('key'), {colname: 'val'})

    def test_static_composite_slicing(self):
        cf = ColumnFamily(pool, 'StaticComposite')
        u1 = uuid.uuid1()
        u4 = uuid.uuid4()
        col0 = (0, 1, u1, u4, '', '', '')
        col1 = (1, 1, u1, u4, '', '', '')
        col2 = (1, 2, u1, u4, '', '', '')
        col3 = (1, 3, u1, u4, '', '', '')
        col4 = (2, 1, u1, u4, '', '', '')
        cf.insert('key2', {col0: '', col1: '', col2: '', col3: '', col4: ''})

        result = cf.get('key2', column_start=((1, True),), column_finish=((1, True),))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(1,), column_finish=((2, False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=((1, True),), column_finish=((2, False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(1, ), column_finish=((2, False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=((0, False), ), column_finish=((2, False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(1, 1), column_finish=(1, 3))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(1, 1), column_finish=(1, (3, True)))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(1, (1, True)), column_finish=((2, False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

    def test_static_composite_get_partial_composite(self):
        cf = ColumnFamily(pool, 'StaticComposite')
        cf.insert('key3', {(123123, 1): 'val'})
        assert_equal(cf.get('key3'), {(123123, 1): 'val'})

    def test_uuid_composites(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'UUIDComposite',
                comparator_type=CompositeType(IntegerType(reversed=True), TimeUUIDType()),
                key_validation_class=TimeUUIDType(),
                default_validation_class=UTF8Type())

        key, u1, u2 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1()
        cf = ColumnFamily(pool, 'UUIDComposite')
        cf.insert(key, {(123123, u1): 'foo'})
        cf.insert(key, {(123123, u1): 'foo', (-1, u2): 'bar', (-123123123, u1): 'baz'})
        assert_equal(cf.get(key), {(123123, u1): 'foo', (-1, u2): 'bar', (-123123123, u1): 'baz'})

    def test_single_component_composite(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'SingleComposite',
                comparator_type=CompositeType(IntegerType()))

        cf = ColumnFamily(pool, 'SingleComposite')
        cf.insert('key', {(123456,): 'val'})
        assert_equal(cf.get('key'), {(123456,): 'val'})

class TestDynamicComposites(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'StaticDynamicComposite',
                                 comparator_type=DynamicCompositeType({'l': LongType(),
                                                                       'i': IntegerType(),
                                                                       'T': TimeUUIDType(reversed=True),
                                                                       'x': LexicalUUIDType(reversed=False),
                                                                       'a': AsciiType(),
                                                                       's': UTF8Type(),
                                                                       'b': BytesType()}))

    @classmethod
    def teardown_class(cls):
        sys = SystemManager()
        sys.drop_column_family(TEST_KS, 'StaticDynamicComposite')

    def setUp(self):
        global a, b, i, I, x, l, t, T, s

        component = namedtuple('DynamicComponent', ['type','value'])
        ascii_alias = component('a', None)
        bytes_alias = component('b', None)
        integer_alias = component('i', None)
        integer_rev_alias = component('I', None)
        lexicaluuid_alias = component('x', None)
        long_alias = component('l', None)
        timeuuid_alias = component('t', None)
        timeuuid_rev_alias = component('T', None)
        utf8_alias = component('s', None)

        _r = lambda t, v: t._replace(value=v) 
        a = lambda v: _r(ascii_alias, v)
        b = lambda v: _r(bytes_alias, v)
        i = lambda v: _r(integer_alias, v)
        I = lambda v: _r(integer_rev_alias, v)
        x = lambda v: _r(lexicaluuid_alias, v)
        l = lambda v: _r(long_alias, v)
        t = lambda v: _r(timeuuid_alias, v)
        T = lambda v: _r(timeuuid_rev_alias, v)
        s = lambda v: _r(utf8_alias, v)

    def test_static_composite_basic(self):
        cf = ColumnFamily(pool, 'StaticDynamicComposite')
        colname = (l(127312831239123123), i(1), T(uuid.uuid1()), x(uuid.uuid4()), a('foo'), s(u'ba\u0254r'), b('baz'))
        cf.insert('key', {colname: 'val'})
        assert_equal(cf.get('key'), {colname: 'val'})

    def test_static_composite_slicing(self):
        cf = ColumnFamily(pool, 'StaticDynamicComposite')
        u1 = uuid.uuid1()
        u4 = uuid.uuid4()
        col0 = (l(0), i(1), T(u1), x(u4), a(''), s(''), b(''))
        col1 = (l(1), i(1), T(u1), x(u4), a(''), s(''), b(''))
        col2 = (l(1), i(2), T(u1), x(u4), a(''), s(''), b(''))
        col3 = (l(1), i(3), T(u1), x(u4), a(''), s(''), b(''))
        col4 = (l(2), i(1), T(u1), x(u4), a(''), s(''), b(''))
        cf.insert('key2', {col0: '', col1: '', col2: '', col3: '', col4: ''})

        result = cf.get('key2', column_start=((l(1), True),), column_finish=((l(1), True),))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(l(1),), column_finish=((l(2), False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=((l(1), True),), column_finish=((l(2), False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(l(1), ), column_finish=((l(2), False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=((l(0), False), ), column_finish=((l(2), False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(l(1), i(1)), column_finish=(l(1), i(3)))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(l(1), i(1)), column_finish=(l(1), (i(3), True)))
        assert_equal(result, {col1: '', col2: '', col3: ''})

        result = cf.get('key2', column_start=(l(1), (i(1), True)), column_finish=((l(2), False), ))
        assert_equal(result, {col1: '', col2: '', col3: ''})

    def test_static_composite_get_partial_composite(self):
        cf = ColumnFamily(pool, 'StaticDynamicComposite')
        cf.insert('key3', {(l(123123), i(1)): 'val'})
        assert_equal(cf.get('key3'), {(l(123123), i(1)): 'val'})

    def test_uuid_composites(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'UUIDDynamicComposite',
                comparator_type=DynamicCompositeType({'I': IntegerType(reversed=True), 't': TimeUUIDType()}),
                key_validation_class=TimeUUIDType(),
                default_validation_class=UTF8Type())

        key, u1, u2 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1()
        cf = ColumnFamily(pool, 'UUIDDynamicComposite')
        cf.insert(key, {(I(123123), t(u1)): 'foo'})
        cf.insert(key, {(I(123123), t(u1)): 'foo', (I(-1), t(u2)): 'bar', (I(-123123123), t(u1)): 'baz'})
        assert_equal(cf.get(key), {(I(123123), t(u1)): 'foo', (I(-1), t(u2)): 'bar', (I(-123123123), t(u1)): 'baz'})

    def test_single_component_composite(self):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'SingleDynamicComposite',
                comparator_type=DynamicCompositeType({'i': IntegerType()}))

        cf = ColumnFamily(pool, 'SingleDynamicComposite')
        cf.insert('key', {(i(123456),): 'val'})
        assert_equal(cf.get('key'), {(i(123456),): 'val'})

class TestBigInt(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'StdInteger', comparator_type=IntegerType())

    @classmethod
    def teardown_class(cls):
        sys = SystemManager()
        sys.drop_column_family(TEST_KS, 'StdInteger')

    def setUp(self):
        self.key = 'TestBigInt'
        self.cf = ColumnFamily(pool, 'StdInteger')

    def tearDown(self):
        self.cf.remove(self.key)

    def test_negative_integers(self):
        self.cf.insert(self.key, {-1: '-1'})
        self.cf.insert(self.key, {-12342390: '-12342390'})
        self.cf.insert(self.key, {-255: '-255'})
        self.cf.insert(self.key, {-256: '-256'})
        self.cf.insert(self.key, {-257: '-257'})
        for key, cols in self.cf.get_range():
            self.assertEquals(str(cols.keys()[0]), cols.values()[0])

class TestTimeUUIDs(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'TestTimeUUIDs', comparator_type=TimeUUIDType())
        sys.close()
        cls.cf_time = ColumnFamily(pool, 'TestTimeUUIDs')

    def test_datetime_to_uuid(self):
        cf_time = TestTimeUUIDs.cf_time
        key = 'key1'
        timeline = []

        timeline.append(datetime.utcnow())
        time1 = uuid1()
        col1 = {time1: '0'}
        cf_time.insert(key, col1)
        time.sleep(1)

        timeline.append(datetime.utcnow())
        time2 = uuid1()
        col2 = {time2: '1'}
        cf_time.insert(key, col2)
        time.sleep(1)

        timeline.append(datetime.utcnow())

        cols = {time1: '0', time2: '1'}

        assert_equal(cf_time.get(key, column_start=timeline[0])                            , cols)
        assert_equal(cf_time.get(key,                           column_finish=timeline[2]) , cols)
        assert_equal(cf_time.get(key, column_start=timeline[0], column_finish=timeline[2]) , cols)
        assert_equal(cf_time.get(key, column_start=timeline[0], column_finish=timeline[2]) , cols)
        assert_equal(cf_time.get(key, column_start=timeline[0], column_finish=timeline[1]) , col1)
        assert_equal(cf_time.get(key, column_start=timeline[1], column_finish=timeline[2]) , col2)
        cf_time.remove(key)

    def test_time_to_uuid(self):
        cf_time = TestTimeUUIDs.cf_time
        key = 'key1'
        timeline = []

        timeline.append(time.time())
        time1 = uuid1()
        col1 = {time1: '0'}
        cf_time.insert(key, col1)
        time.sleep(0.1)

        timeline.append(time.time())
        time2 = uuid1()
        col2 = {time2: '1'}
        cf_time.insert(key, col2)
        time.sleep(0.1)

        timeline.append(time.time())

        cols = {time1:'0', time2: '1'}

        assert_equal(cf_time.get(key, column_start=timeline[0])                            , cols)
        assert_equal(cf_time.get(key,                           column_finish=timeline[2]) , cols)
        assert_equal(cf_time.get(key, column_start=timeline[0], column_finish=timeline[2]) , cols)
        assert_equal(cf_time.get(key, column_start=timeline[0], column_finish=timeline[2]) , cols)
        assert_equal(cf_time.get(key, column_start=timeline[0], column_finish=timeline[1]) , col1)
        assert_equal(cf_time.get(key, column_start=timeline[1], column_finish=timeline[2]) , col2)
        cf_time.remove(key)

    def test_auto_time_to_uuid1(self):
        cf_time = TestTimeUUIDs.cf_time
        key = 'key1'
        t = time.time()
        col = {t: 'foo'}
        cf_time.insert(key, col)
        uuid_res = cf_time.get(key).keys()[0]
        timestamp = convert_uuid_to_time(uuid_res)
        assert_almost_equal(timestamp, t, places=3)
        cf_time.remove(key)

class TestTypeErrors(unittest.TestCase):

    def test_packing_enabled(self):
        self.cf = ColumnFamily(pool, 'Standard1')
        self.cf.insert('key', {'col': 'val'})
        assert_raises(TypeError, self.cf.insert, args=('key', {123: 'val'}))
        assert_raises(TypeError, self.cf.insert, args=('key', {'col': 123}))
        assert_raises(TypeError, self.cf.insert, args=('key', {123: 123}))
        self.cf.remove('key')

    def test_packing_disabled(self):
        self.cf = ColumnFamily(pool, 'Standard1', autopack_names=False, autopack_values=False)
        self.cf.insert('key', {'col': 'val'})
        assert_raises(TypeError, self.cf.insert, args=('key', {123: 'val'}))
        assert_raises(TypeError, self.cf.insert, args=('key', {'col': 123}))
        assert_raises(TypeError, self.cf.insert, args=('key', {123: 123}))
        self.cf.remove('key')

class TestDateTypes(unittest.TestCase):

    def _compare_dates(self, d1, d2):
        self.assertEquals(d1.timetuple(), d2.timetuple())
        self.assertEquals(int(d1.microsecond/1e3), int(d2.microsecond/1e3))

    def test_compatibility(self):
        self.cf = ColumnFamily(pool, 'Standard1')
        self.cf.column_validators['date'] = OldPycassaDateType()

        d = datetime.utcnow()
        self.cf.insert('key1', {'date': d})
        self._compare_dates(self.cf.get('key1')['date'], d)

        self.cf.column_validators['date'] = IntermediateDateType()
        self._compare_dates(self.cf.get('key1')['date'], d)
        self.cf.insert('key1', {'date': d})
        self._compare_dates(self.cf.get('key1')['date'], d)

        self.cf.column_validators['date'] = DateType()
        self._compare_dates(self.cf.get('key1')['date'], d)
        self.cf.insert('key1', {'date': d})
        self._compare_dates(self.cf.get('key1')['date'], d)
        self.cf.remove('key1')

class TestPackerOverride(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(TEST_KS, 'CompositeOverrideCF',
                comparator_type=CompositeType(AsciiType(), AsciiType()),
                default_validation_class=AsciiType())

    @classmethod
    def teardown_class(cls):
        sys = SystemManager()
        sys.drop_column_family(TEST_KS, 'CompositeOverrideCF')

    def test_column_validator(self):
        cf = ColumnFamily(pool, 'CompositeOverrideCF')
        cf.column_validators[('a', 'b')] = BooleanType()
        cf.insert('key', {('a', 'a'): 'foo', ('a', 'b'): True})
        assert_equal(cf.get('key'), {('a', 'a'): 'foo', ('a', 'b'): True})

        assert_equal(cf.column_validators[('a', 'b')].__class__, BooleanType)

        keys = cf.column_validators.keys()
        assert_equal(keys, [('a', 'b')])

        del cf.column_validators[('a', 'b')]
        assert_raises(KeyError, cf.column_validators.__getitem__, ('a', 'b'))

class TestCustomTypes(unittest.TestCase):

    class IntString(CassandraType):

        @staticmethod
        def pack(intval):
            return str(intval)

        @staticmethod
        def unpack(strval):
            return int(strval)

    class IntString2(CassandraType):

        def __init__(self, *args, **kwargs):
            self.pack = lambda val: str(val)
            self.unpack = lambda val: int(val)

    def test_staticmethod_funcs(self):
        self.cf = ColumnFamily(pool, 'Standard1')
        self.cf.key_validation_class = TestCustomTypes.IntString()
        self.cf.insert(1234, {'col': 'val'})
        assert_equal(self.cf.get(1234), {'col': 'val'})

    def test_constructor_lambdas(self):
        self.cf = ColumnFamily(pool, 'Standard1')
        self.cf.key_validation_class = TestCustomTypes.IntString2()
        self.cf.insert(1234, {'col': 'val'})
        assert_equal(self.cf.get(1234), {'col': 'val'})

class TestCustomComposite(unittest.TestCase):
    """
    Test CompositeTypes with custom inner types.
    """

    # Some contrived scenarios
    class IntDateType(CassandraType):
        """
        Represent a date as an integer. E.g.: March 05, 2012 = 20120305
        """
        @staticmethod
        def pack(v, *args, **kwargs):
            assert type(v) in (datetime, date), "Invalid arg"
            str_date = v.strftime("%Y%m%d")
            return marshal.encode_int(int(str_date))

        @staticmethod
        def unpack(v, *args, **kwargs):
            int_date = marshal.decode_int(v)
            return date(*time.strptime(str(int_date), "%Y%m%d")[0:3])

    class IntString(CassandraType):

        @staticmethod
        def pack(intval):
            return str(intval)

        @staticmethod
        def unpack(strval):
            return int(strval)

    class IntString2(CassandraType):

        def __init__(self, *args, **kwargs):
            self.pack = lambda val: str(val)
            self.unpack = lambda val: int(val)

    @classmethod
    def setup_class(cls):
        sys = SystemManager()
        sys.create_column_family(
            TEST_KS,
            'CustomComposite1',
            comparator_type=CompositeType(
                IntegerType(),
                UTF8Type()))

    @classmethod
    def teardown_class(cls):
        sys = SystemManager()
        sys.drop_column_family(TEST_KS, 'CustomComposite1')

    def test_static_composite_basic(self):
        cf = ColumnFamily(pool, 'CustomComposite1')
        colname = (20120305, '12345')
        cf.insert('key', {colname: 'val1'})
        assert_equal(cf.get('key'), {colname: 'val1'})

    def test_insert_with_custom_composite(self):
        cf_std = ColumnFamily(pool, 'CustomComposite1')
        cf_cust = ColumnFamily(pool, 'CustomComposite1')
        cf_cust.column_name_class = CompositeType(
                TestCustomComposite.IntDateType(),
                TestCustomComposite.IntString())

        std_col = (20120311, '321')
        cust_col = (date(2012, 3, 11), 321)
        cf_cust.insert('cust_insert_key_1', {cust_col: 'cust_insert_val_1'})
        assert_equal(cf_std.get('cust_insert_key_1'),
                {std_col: 'cust_insert_val_1'})

    def test_retrieve_with_custom_composite(self):
        cf_std = ColumnFamily(pool, 'CustomComposite1')
        cf_cust = ColumnFamily(pool, 'CustomComposite1')
        cf_cust.column_name_class = CompositeType(
                TestCustomComposite.IntDateType(),
                TestCustomComposite.IntString())

        std_col = (20120312, '321')
        cust_col = (date(2012, 3, 12), 321)
        cf_std.insert('cust_insert_key_2', {std_col: 'cust_insert_val_2'})
        assert_equal(cf_cust.get('cust_insert_key_2'),
                {cust_col: 'cust_insert_val_2'})

    def test_composite_slicing(self):
        cf_std = ColumnFamily(pool, 'CustomComposite1')
        cf_cust = ColumnFamily(pool, 'CustomComposite1')
        cf_cust.column_name_class = CompositeType(
                TestCustomComposite.IntDateType(),
                TestCustomComposite.IntString2())

        col0 = (20120101, '123')
        col1 = (20120102, '123')
        col2 = (20120102, '456')
        col3 = (20120102, '789')
        col4 = (20120103, '123')

        dt0 = date(2012, 1, 1)
        dt1 = date(2012, 1, 2)
        dt2 = date(2012, 1, 3)

        col1_cust = (dt1, 123)
        col2_cust = (dt1, 456)
        col3_cust = (dt1, 789)

        cf_std.insert('key2', {col0: '', col1: '', col2: '', col3: '', col4: ''})

        def check(column_start, column_finish, col_reversed=False):
            result = cf_cust.get('key2', column_start=column_start,
                    column_finish=column_finish, column_reversed=col_reversed)

            assert_equal(result, {col1_cust: '', col2_cust: '', col3_cust: ''})

        # Defaults should be inclusive on both ends
        check((dt1,), (dt1,))
        check((dt1,), (dt1,), True)

        check(((dt1, True),), ((dt1, True),))
        check((dt1,), ((dt2, False),))
        check(((dt1, True),), ((dt2, False),))
        check(((dt0, False),), ((dt2, False),))

        check((dt1, 123), (dt1, 789))
        check((dt1, 123), (dt1, (789, True)))
        check((dt1, (123, True)), ((dt2, False),))

        # Test inclusive ends for reversed
        check(((dt1, True),), ((dt1, True),), True)
        check( (dt1,),        ((dt1, True),), True)
        check(((dt1, True),),  (dt1,),        True)

        # Test exclusive ends for reversed
        check(((dt2, False),), ((dt0, False),), True)
        check(((dt2, False),),  (dt1,),         True)
        check((dt1,),          ((dt0, False),), True)

